import datetime
from collections import defaultdict

import mock
import numpy as np
import pandas as pd
import pytest
import pytz
import responses
from freezegun import freeze_time
from responses import matchers

from gantry.query.core import queryframe
from gantry.query.core.queryframe import _get_metadata

from ..conftest import END_TIME, ORIGIN, START_TIME


@pytest.fixture
def query_frame_obj(api_client_obj, query_info_obj):
    return queryframe.GantryQueryFrame(api_client_obj, query_info_obj)


def test_get_base_query_params(query_info_obj):
    assert query_info_obj.get_base_query_params() == {
        "start_time": START_TIME,
        "end_time": END_TIME,
        "version": "1.2.3",
    }


@pytest.mark.parametrize(
    ["start_time", "end_time", "expected"],
    [
        (START_TIME, END_TIME, (START_TIME, END_TIME)),
        (
            "August 14, 2015 EST",
            "July 4, 2013 PST",
            (
                datetime.datetime(2015, 8, 14, 5, 0, tzinfo=pytz.UTC),
                datetime.datetime(2013, 7, 4, 8, 0, tzinfo=pytz.UTC),
            ),
        ),
    ],
)
def test_resolve_time_strings(start_time, end_time, expected):
    assert queryframe.GantryQueryFrame.resolve_time_strings(start_time, end_time) == expected


@pytest.mark.parametrize(
    ["start_time", "end_time"],
    [
        ("abc", END_TIME),
        (START_TIME, "foobar"),
    ],
)
def test_resolve_time_strings_error(start_time, end_time):
    with pytest.raises(ValueError):
        queryframe.GantryQueryFrame.resolve_time_strings(start_time, end_time)


@pytest.mark.parametrize(
    ["tags", "env", "expected"],
    [
        (None, None, None),
        ({"foo": "bar"}, None, {"foo": "bar"}),
        ({"foo": "bar"}, "dev", {"foo": "bar", "env": "dev"}),
        ({"foo": "bar", "env": "another"}, "dev", {"foo": "bar", "env": "dev"}),
        ({"foo": "bar", "env": "another"}, None, {"foo": "bar", "env": "another"}),
        (None, "dev", {"env": "dev"}),
    ],
)
@pytest.mark.parametrize("n", [None, 1, 5])
def test_populate_raw_request_descending_no_pagination(n, tags, env, expected, api_client_obj):
    query_info_obj = queryframe.QueryInfo(
        "af0d0317-adbf-4524-b838-9dddf1c43fc8", "foobar", "1.2.3", env, START_TIME, END_TIME
    )
    query_frame_obj = queryframe.GantryQueryFrame(api_client_obj, query_info_obj)

    matcher = {"tags": expected} if expected else {}

    if n:
        matcher["limit"] = n

    matcher["order"] = "descending"
    matcher["include_join_id"] = True
    expected_events = list("these-are-events")[:n] if n else list("these-are-events")
    with responses.RequestsMock() as rsps:
        rsps.add(
            responses.POST,
            "{}/api/v1/applications/af0d0317-adbf-4524-b838-9dddf1c43fc8/raw_data".format(ORIGIN),
            json={
                "response": "ok",
                "events": expected_events[::-1],
                "columns": "these-are-columns",
                "metadata": {
                    "count": n or 1,
                    "has_next": False,
                },
            },
            match=[matchers.json_params_matcher(matcher)],
        )

        assert (
            query_frame_obj._populate_raw_request(query_filters=[], ascending=False, n=n, tags=tags)
            == expected_events
        )


@pytest.mark.parametrize(
    ["tags", "env", "expected"],
    [
        (None, None, None),
        ({"foo": "bar"}, None, {"foo": "bar"}),
        ({"foo": "bar"}, "dev", {"foo": "bar", "env": "dev"}),
        ({"foo": "bar", "env": "another"}, "dev", {"foo": "bar", "env": "dev"}),
        ({"foo": "bar", "env": "another"}, None, {"foo": "bar", "env": "another"}),
        (None, "dev", {"env": "dev"}),
    ],
)
@pytest.mark.parametrize("n", [None, 1, 5])
def test_populate_raw_request_ascending_no_pagination(n, tags, env, expected, api_client_obj):
    query_info_obj = queryframe.QueryInfo(
        "af0d0317-adbf-4524-b838-9dddf1c43fc8", "foobar", "1.2.3", env, START_TIME, END_TIME
    )
    query_frame_obj = queryframe.GantryQueryFrame(api_client_obj, query_info_obj)

    matcher = {"tags": expected} if expected else {}

    if n:
        matcher["limit"] = n

    matcher["order"] = "ascending"
    matcher["include_join_id"] = True

    expected_events = list("these-are-events")[:n] if n else list("these-are-events")
    with responses.RequestsMock() as rsps:
        rsps.add(
            responses.POST,
            "{}/api/v1/applications/af0d0317-adbf-4524-b838-9dddf1c43fc8/raw_data".format(ORIGIN),
            json={
                "response": "ok",
                "events": expected_events,
                "columns": "these-are-columns",
                "metadata": {
                    "count": n or 1,
                    "has_next": False,
                },
            },
            match=[matchers.json_params_matcher(matcher)],
        )

        assert (
            query_frame_obj._populate_raw_request(query_filters=[], ascending=True, n=n, tags=tags)
            == expected_events
        )


@pytest.mark.parametrize(
    ["tags", "env", "expected"],
    [
        (None, None, None),
        ({"foo": "bar"}, None, {"foo": "bar"}),
        ({"foo": "bar"}, "dev", {"foo": "bar", "env": "dev"}),
        ({"foo": "bar", "env": "another"}, "dev", {"foo": "bar", "env": "dev"}),
        ({"foo": "bar", "env": "another"}, None, {"foo": "bar", "env": "another"}),
        (None, "dev", {"env": "dev"}),
    ],
)
@pytest.mark.parametrize("n", [10, 53, 101])
def test_populate_raw_request_ascending_pagination(n, tags, env, expected, api_client_obj):
    query_info_obj = queryframe.QueryInfo(
        "af0d0317-adbf-4524-b838-9dddf1c43fc8", "foobar", "1.2.3", env, START_TIME, END_TIME
    )
    query_frame_obj = queryframe.GantryQueryFrame(api_client_obj, query_info_obj)

    total_num_events = 100
    all_events = list(range(total_num_events))
    requested_events = all_events[:n]
    up_to = n - (n // 2)  # return less than n to test pagination

    # Two matchers, one for the first page and one for the second
    first_matcher = {"tags": expected} if expected else {}
    first_matcher["limit"] = n
    first_matcher["order"] = "ascending"
    first_matcher["include_join_id"] = True

    second_matcher = first_matcher.copy()
    second_matcher["offset"] = len(requested_events[:up_to])

    # Two responses, one for the first page and one for the second
    first_expected_response = {
        "response": "ok",
        "events": requested_events[:up_to],
        "columns": "these-are-columns",
        "metadata": {
            "count": total_num_events,
            "has_next": True,
        },
    }
    second_expected_response = {
        "response": "ok",
        "events": requested_events[up_to:],
        "columns": "these-are-columns",
        "metadata": {
            "count": total_num_events,
            "has_next": False,
        },
    }
    url = "{}/api/v1/applications/af0d0317-adbf-4524-b838-9dddf1c43fc8/raw_data".format(ORIGIN)

    # pagination means we need to make multiple POST requests
    with responses.RequestsMock() as rsps:
        rsps.add(
            responses.POST,
            url,
            json=first_expected_response,
            match=[matchers.json_params_matcher(first_matcher)],
        )
        rsps.add(
            responses.POST,
            url,
            json=second_expected_response,
            match=[matchers.json_params_matcher(second_matcher)],
        )
        assert (
            query_frame_obj._populate_raw_request(query_filters=[], ascending=True, n=n, tags=tags)
            == requested_events
        )


@pytest.mark.parametrize(
    ["tags", "env", "expected"],
    [
        (None, None, None),
        ({"foo": "bar"}, None, {"foo": "bar"}),
        ({"foo": "bar"}, "dev", {"foo": "bar", "env": "dev"}),
        ({"foo": "bar", "env": "another"}, "dev", {"foo": "bar", "env": "dev"}),
        ({"foo": "bar", "env": "another"}, None, {"foo": "bar", "env": "another"}),
        (None, "dev", {"env": "dev"}),
    ],
)
@pytest.mark.parametrize("n", [10, 53, 101])
def test_populate_raw_request_descending_pagination(n, tags, env, expected, api_client_obj):
    query_info_obj = queryframe.QueryInfo(
        "af0d0317-adbf-4524-b838-9dddf1c43fc8", "foobar", "1.2.3", env, START_TIME, END_TIME
    )
    query_frame_obj = queryframe.GantryQueryFrame(api_client_obj, query_info_obj)

    total_num_events = 100
    all_events = list(range(total_num_events))
    requested_events = all_events[:n]
    up_to = n - (n // 2)  # return less than n to test pagination

    # Two matchers, one for the first page and one for the second
    first_matcher = {"tags": expected} if expected else {}
    first_matcher["limit"] = n
    first_matcher["order"] = "descending"
    first_matcher["include_join_id"] = True

    second_matcher = first_matcher.copy()
    second_matcher["offset"] = len(requested_events[:up_to])

    # Two responses, one for the first page and one for the second
    first_expected_response = {
        "response": "ok",
        "events": requested_events[::-1][:up_to],
        "columns": "these-are-columns",
        "metadata": {
            "count": total_num_events,
            "has_next": True,
        },
    }
    second_expected_response = {
        "response": "ok",
        "events": requested_events[::-1][up_to:],
        "columns": "these-are-columns",
        "metadata": {
            "count": total_num_events,
            "has_next": False,
        },
    }
    url = "{}/api/v1/applications/af0d0317-adbf-4524-b838-9dddf1c43fc8/raw_data".format(ORIGIN)

    # pagination means we need to make multiple POST requests
    with responses.RequestsMock() as rsps:
        rsps.add(
            responses.POST,
            url,
            json=first_expected_response,
            match=[matchers.json_params_matcher(first_matcher)],
        )
        rsps.add(
            responses.POST,
            url,
            json=second_expected_response,
            match=[matchers.json_params_matcher(second_matcher)],
        )
        assert (
            query_frame_obj._populate_raw_request(query_filters=[], ascending=False, n=n, tags=tags)
            == requested_events
        )


@pytest.mark.parametrize("provided_tags", [None, {}, {"foo": "bar"}])
@pytest.mark.parametrize("num_events", [1, 5, 10, 20])
@pytest.mark.parametrize("column", ["A", None, "not_a_column"])
@pytest.mark.parametrize("method", ["_head", "_tail"])
@mock.patch("gantry.query.core.queryframe.GantryQueryFrame._populate_raw_request")
@mock.patch("gantry.query.core.dataframe.GantryQueryFrame._dtypes", new_callable=mock.PropertyMock)
def test_head_tail(
    mock_dtypes, mock_populate_raw, method, column, num_events, provided_tags, query_frame_obj
):
    columns = ["A", "B", "C"]

    def fake_populate_raw(n, ascending, query_filters, tags):
        assert provided_tags == tags
        events = defaultdict(list)
        for c in columns:
            if c != "not_a_column":
                events[c] = list(range(num_events))
        return events

    mock_populate_raw.side_effect = fake_populate_raw
    mock_dtypes.return_value = {c: "int" for c in column} if column else {}

    if column == "not_a_column":
        with pytest.raises(ValueError):
            _ = getattr(query_frame_obj, method)(n=num_events, column=column, tags=provided_tags)

    else:
        df = getattr(query_frame_obj, method)(n=num_events, column=column, tags=provided_tags)

        expected = []
        for i in range(num_events):
            if column:
                expected.append({"A": i})
            else:
                expected.append({"A": i, "B": i, "C": i})

        pd.testing.assert_frame_equal(df, pd.DataFrame(expected))


@pytest.mark.parametrize("provided_tags", [None, {}, {"foo": "bar"}])
@pytest.mark.parametrize("num_events", [1, 5, 10, 20])
@pytest.mark.parametrize("column", ["A", None, "not_a_column"])
@mock.patch("gantry.query.core.queryframe.GantryQueryFrame._populate_raw_request")
@mock.patch("gantry.query.core.dataframe.GantryQueryFrame._dtypes", new_callable=mock.PropertyMock)
def test_fetch(mock_dtypes, mock_populate_raw, column, num_events, provided_tags, query_frame_obj):
    columns = ["A", "B", "C"]

    def fake_populate_raw(n, ascending, query_filters, tags):
        assert tags == provided_tags
        events = defaultdict(list)
        for c in columns:
            if c != "not_a_column":
                events[c] = list(range(num_events))
        return events

    mock_populate_raw.side_effect = fake_populate_raw
    mock_dtypes.return_value = {c: "int" for c in column} if column else {}

    if column == "not_a_column":
        with pytest.raises(ValueError):
            _ = query_frame_obj._fetch(column=column, tags=provided_tags)

    else:
        df = query_frame_obj._fetch(column=column, tags=provided_tags)

        expected = []
        for i in range(num_events):
            if column:
                expected.append({"A": i})
            else:
                expected.append({"A": i, "B": i, "C": i})

        pd.testing.assert_frame_equal(df, pd.DataFrame(expected))


@pytest.mark.parametrize("method", ["_head", "_tail"])
def test_head_tail_invalid_param(method, query_frame_obj):
    with pytest.raises(ValueError):
        _ = getattr(query_frame_obj, method)(n=0)


@pytest.mark.parametrize("method", ["_fetch"])
def test_fetch_invalid_param(method, query_frame_obj):
    with pytest.raises(TypeError):
        # _fetch takes no parameters
        _ = getattr(query_frame_obj, method)(n=0)


@freeze_time("2020-01-01")
@mock.patch("gantry.query.core.dataframe.GantryQueryFrame._dtypes", new_callable=mock.PropertyMock)
def test_prepare_pandas_dataframe_with_timestamp(mock_dtypes, query_frame_obj):
    _dtypes = {
        "A": "int",
        "B": "int",
        "C": "str",
    }
    mock_dtypes.return_value = _dtypes

    # Compute time in ms since epoch (1/1/1970), create events
    current_unix_time = int(
        (datetime.datetime.now() - datetime.datetime(1970, 1, 1)).total_seconds() * 1000
    )
    events = [
        {"__time": current_unix_time, "A": 1, "B": 2, "C": "foo"},
        {"__time": current_unix_time + int(1e3), "A": 23, "B": 52, "C": "bar"},
        {"__time": current_unix_time, "A": 4, "B": 2, "C": "baz"},
        {"__time": current_unix_time + int(1e5), "A": 5, "B": 2, "C": "goof"},
    ]
    df = query_frame_obj._prepare_pandas_dataframe(events)

    # Check that the index is a DatetimeIndex
    assert isinstance(df.index, pd.DatetimeIndex)
    assert df.index.name == "timestamp"

    # Check that the index is sorted
    correct_times = [
        current_unix_time,
        current_unix_time,
        current_unix_time + int(1e3),
        current_unix_time + int(1e5),
    ]

    assert all(
        [
            df.index.values[i].astype("datetime64[ms]").astype("int") == correct_times[i]
            for i in range(len(df))
        ]
    )


@mock.patch("gantry.query.core.dataframe.GantryQueryFrame._dtypes", new_callable=mock.PropertyMock)
def test_prepare_pandas_dataframe_without_timestamp(mock_dtypes, query_frame_obj):
    _dtypes = {
        "A": "bool",
        "B": "category",
        "C": "float",
        "D": "int",
        "E": "uuid",
        "F": "str",
        "G_some_none": "bool",
        "H_all_none": "str",
    }
    mock_dtypes.return_value = _dtypes

    events = [
        {
            "A": "true",
            "B": "foo",
            "C": 1.0,
            "D": 1,
            "E": "asd123",
            "F": "foo",
            "G_some_none": "true",
            "H_all_none": None,
        },
        {
            "A": "false",
            "B": "bar",
            "C": 2.0,
            "D": 2,
            "E": "345436gdfkjdn",
            "F": "bar",
            "G_some_none": None,
            "H_all_none": None,
        },
        {
            "A": "true",
            "B": "baz",
            "C": 3.0,
            "D": 3,
            "E": "sadfs3456",
            "F": "baz",
            "G_some_none": "false",
            "H_all_none": None,
        },
    ]

    df = query_frame_obj._prepare_pandas_dataframe(events)

    expected_dtypes = {
        "A": "boolean",
        "B": "category",
        "C": "float",
        "D": "int",
        "E": "object",
        "F": "string",
        "G_some_none": "boolean",
        "H_all_none": "string",
    }

    # ordered alphabetically
    assert df.columns.tolist() == list(expected_dtypes.keys())

    for actual, expected in zip(df.dtypes.tolist(), list(expected_dtypes.values())):
        assert actual == expected

    # all None column is bool but values are "not available"
    assert isinstance(df["H_all_none"].unique()[0], type(pd.NA))

    # Check that DataFrame.index is the typical RangeIndex counter
    # Contrasts the case where the index is a DatetimeIndex
    assert isinstance(df.index, pd.RangeIndex)


@pytest.mark.parametrize("column", [None, "A"])
@mock.patch("gantry.query.core.dataframe.GantryQueryFrame._dtypes", new_callable=mock.PropertyMock)
def test_prepare_pandas_dataframe_empty(mock_dtypes, column, query_frame_obj):
    _dtypes = {
        "A": "bool",
        "B": "category",
        "C": "float",
    }
    mock_dtypes.return_value = pd.Series(_dtypes)
    df = query_frame_obj._prepare_pandas_dataframe([], column=column)
    assert df.empty

    assert len(df.columns) == 3

    assert df.columns[0] == "A"
    assert df.columns[1] == "B"
    assert df.columns[2] == "C"

    assert df.dtypes[0] == "bool"
    assert df.dtypes[1] == "category"
    assert df.dtypes[2] == np.dtype("float64")


def test_get_metadata(query_frame_obj):
    application_name = "test name"
    query_params = frozenset(
        {
            "start_time": START_TIME,
            "end_time": END_TIME,
            "version": "1.2.3",
        }.items()
    )
    with responses.RequestsMock() as rsps:
        rsps.add(
            responses.GET,
            "{}/api/v1/models/{}/schemas".format(ORIGIN, application_name),
            json={
                "response": "ok",
                "data": {"some": "data"},
            },
        )

        assert _get_metadata(query_frame_obj.api_client, application_name, query_params) == {
            "some": "data"
        }
        assert _get_metadata(query_frame_obj.api_client, application_name, query_params) == {
            "some": "data"
        }  # @cached will protect us for the second request, no request mock needed


def test_metadata(query_frame_obj):
    with responses.RequestsMock() as rsps:
        rsps.add(
            responses.GET,
            "{}/api/v1/models/foobar/schemas".format(ORIGIN),
            json={"response": "ok", "data": "this-is-the-schema"},
            match=[
                matchers.query_param_matcher(
                    {
                        "start_time": str(START_TIME),
                        "end_time": str(END_TIME),
                        "version": "1.2.3",
                    }
                )
            ],
        )

        assert query_frame_obj.metadata == "this-is-the-schema"


@pytest.mark.parametrize(
    ["metadata", "expected"],
    [
        ({}, {}),
        (
            {
                "feedback_datanodes": [
                    {
                        "dtype": "bool",
                        "event_type": "feedback",
                        "name": "feedback.is_fraud",
                        "node_metadata": {
                            "content_retrieval": ["feedback", "is_fraud"],
                            "dtype": {"id_str": False, "is_categorical": True, "long_str": False},
                        },
                        "node_type": "identity",
                    }
                ],
                "prediction_datanodes": [
                    {
                        "dtype": "float",
                        "event_type": "prediction",
                        "name": "inputs.amt",
                        "node_metadata": {
                            "content_retrieval": ["inputs", "amt"],
                            "dtype": {"id_str": False, "is_categorical": False, "long_str": False},
                        },
                        "node_type": "identity",
                    },
                    {
                        "dtype": "bool",
                        "event_type": "prediction",
                        "name": "outputs.pred_fraud",
                        "node_metadata": {
                            "content_retrieval": ["outputs", "pred_fraud"],
                            "dtype": {"id_str": False, "is_categorical": True, "long_str": False},
                        },
                        "node_type": "identity",
                    },
                    {
                        "dtype": "str",
                        "event_type": "prediction",
                        "name": "inputs.zip",
                        "node_metadata": {
                            "content_retrieval": ["inputs", "zip"],
                            "dtype": {"is_categorical": False, "long_str": True},
                        },
                        "node_type": "identity",
                    },
                    {
                        "dtype": "str",
                        "event_type": "prediction",
                        "name": "inputs.cc_num",
                        "node_metadata": {
                            "content_retrieval": ["inputs", "cc_num"],
                            "dtype": {"is_categorical": False, "long_str": True},
                        },
                        "node_type": "identity",
                    },
                ],
                "projection_datanodes": [
                    {
                        "dtype": "float",
                        "event_type": "prediction",
                        "name": "inputs.category.CommonWordProjection",
                        "node_metadata": {
                            "dtype": {},
                            "projection_class": "CommonWordProjection",
                            "projection_init": {},
                        },
                        "node_type": "projection",
                    }
                ],
            },
            {
                "feedback.is_fraud": "bool",
                "inputs.amt": "float",
                "inputs.category.CommonWordProjection": "float",
                "inputs.cc_num": "str",
                "inputs.zip": "str",
                "outputs.pred_fraud": "bool",
            },
        ),
    ],
)
@mock.patch(
    "gantry.query.core.queryframe.GantryQueryFrame.metadata",
    new_callable=mock.PropertyMock,
)
def test_dtypes(mock_metadata, metadata, expected, query_frame_obj):
    mock_metadata.return_value = metadata
    assert query_frame_obj._dtypes == expected


def test_get_raw_data(query_frame_obj):
    with responses.RequestsMock() as rsps:
        rsps.add(
            responses.POST,
            "{}/api/v1/applications/af0d0317-adbf-4524-b838-9dddf1c43fc8/raw_data".format(ORIGIN),
            json="OK",
            match=[matchers.json_params_matcher({"foo": "bar"})],
        )

        assert (
            queryframe._get_raw_data(
                query_frame_obj.api_client, query_frame_obj.query_info, '{"foo": "bar"}'
            )
            == "OK"
        )



import json

import mock
import pytest
import responses
from click.testing import CliRunner
from responses import matchers

from gantry.cli.data_connector import (
    create,
    delete,
    delete_pipeline,
    list,
    list_pipeline_operations,
    list_pipelines,
)

UNITTEST_HOST = "http://unittest"
DATA_CONNECTOR_NAME = "test_data_connector"
DATA_CONNECTOR_1_ID = "0db220e0-f568-4295-be48-67095fa57f34"
DATA_CONNECTOR_2_ID = "9e09947e-fb18-4ce3-bd17-7c386a3c9980"
DATA_CONNECTOR_ENABLED_STATUS_CMD = "enabled"
DATA_CONNECTOR_ENABLED_STATUS = "True"
DATABASE_NAME = "test_database"
SUPPORTED_CONNECTION_TYPES = ["SNOWFLAKE", "S3"]
DESCRIPTION = "Data connector to log records from my snowflake database"
SECRET_NAME = "test_secret"
APP_NAME = "test_app"
METADATA_ID = "09a26daf-7edb-4e94-aa57-31315ae1a31a"
PIPELINE_NAME = "test_pipeline"
PIPELINE_OPERATION_STATUS = "QUEUED"
PIPELINE_OPERATION_ID = "b4bc71af-37b7-4460-b66c-7ef078949c41"
PIPELINE_ID = "b4bc71af-37b7-4460-b66c-7ef078949c41"
PIPELINE_CONSUMER_ID = "3318762f-a684-480c-b418-d5adcdcb5dc6"
APP_VERSION = "1.0.0"

SNOWFLAKE_OPTIONS = '{"snowflake_schema_name": "PUBLIC","snowflake_table_name": "GANTRY_EVENTS"}'
S3_OPTIONS = (
    '{"s3_bucket_name": "test-bucket-name","s3_filetype": "csv","s3_path_prefix": "test-prefix/"}'
)


@pytest.mark.parametrize(
    ("connection_type", "options"),
    [
        ("SNOWFLAKE", SNOWFLAKE_OPTIONS),
        ("S3", S3_OPTIONS),
    ],
)
def test_create_data_connector(connection_type, options):
    runner = CliRunner()
    cli_args = [
        "--name",
        DATA_CONNECTOR_NAME,
        "--database-name",
        DATABASE_NAME,
        "--connection-type",
        connection_type,
        "--secret-name",
        SECRET_NAME,
        "--description",
        DESCRIPTION,
        "--options",
        options,
    ]

    with responses.RequestsMock() as resp:
        resp.add(
            resp.POST,
            f"{UNITTEST_HOST}/api/v1/data-connectors/sources",
            json={
                "response": "ok",
                "data": {
                    "connection_type": connection_type,
                    "created_at": "2023-01-18T00:18:01.000000",
                    "database_name": DATABASE_NAME,
                    "description": DESCRIPTION,
                    "id": DATA_CONNECTOR_1_ID,
                    "name": DATABASE_NAME,
                    "options": options,
                    "secret_name": SECRET_NAME,
                    "type": "SOURCE",
                    "updated_at": "2023-01-18T00:18:01.000000",
                },
            },
            headers={"Content-Type": "application/json"},
            match=[
                matchers.json_params_matcher(
                    {
                        "name": DATA_CONNECTOR_NAME,
                        "database_name": DATABASE_NAME,
                        "connection_type": connection_type,
                        "secret_name": SECRET_NAME,
                        "description": DESCRIPTION,
                        "options": json.loads(options),
                    }
                )
            ],
        )

        result = runner.invoke(
            create,
            cli_args,
            env={"GANTRY_API_KEY": "test", "GANTRY_LOGS_LOCATION": UNITTEST_HOST},
        )

        assert result.exit_code == 0
        assert "--> A source data connector has been created" in result.output


def test_create_data_connector_input_validation_failure():
    """
    Tests that invalid connection type throws an error
    """
    runner = CliRunner()
    cli_args = [
        "--name",
        DATA_CONNECTOR_NAME,
        "--connection-type",
        "NOT_VALID_TYPE",
        "--secret-name",
        SECRET_NAME,
        "--description",
        DESCRIPTION,
        "--options",
        S3_OPTIONS,
    ]

    result = runner.invoke(
        create,
        cli_args,
        env={"GANTRY_API_KEY": "test", "GANTRY_LOGS_LOCATION": UNITTEST_HOST},
    )

    assert result.exit_code == 2
    assert "Error: Invalid value for '--connection-type': 'NOT_VALID_TYPE'" in result.output


@pytest.mark.parametrize("connection_type", SUPPORTED_CONNECTION_TYPES)
@mock.patch("gantry.cli.data_connector.APIClient.request")
def test_create_data_connector_api_failure(mock_api_client: mock.Mock, connection_type):
    """
    Tests that invalid connection type throws an error
    """
    mock_api_client.return_value = {"response": "error", "error": "some error"}

    runner = CliRunner()
    cli_args = [
        "--name",
        DATA_CONNECTOR_NAME,
        "--database-name",
        DATABASE_NAME,
        "--connection-type",
        connection_type,
        "--secret-name",
        SECRET_NAME,
        "--description",
        DESCRIPTION,
        "--options",
        SNOWFLAKE_OPTIONS,
    ]

    result = runner.invoke(
        create,
        cli_args,
        env={"GANTRY_API_KEY": "test", "GANTRY_LOGS_LOCATION": UNITTEST_HOST},
    )

    assert result.exit_code == 1


def test_list_data_connectors():
    """
    Tests that correct message displays when list command returns data connectors
    """
    runner = CliRunner()

    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{UNITTEST_HOST}/api/v1/data-connectors",
            json={
                "response": "ok",
                "data": [
                    {
                        "connection_type": "SNOWFLAKE",
                        "created_at": "2023-01-18T00:18:01.000000",
                        "database_name": DATABASE_NAME,
                        "description": DESCRIPTION,
                        "id": DATA_CONNECTOR_1_ID,
                        "name": DATABASE_NAME,
                        "options": SNOWFLAKE_OPTIONS,
                        "secret_name": SECRET_NAME,
                        "type": "SOURCE",
                        "updated_at": "2023-01-18T00:18:01.000000",
                    },
                    {
                        "connection_type": "S3",
                        "created_at": "2023-01-18T00:18:01.000000",
                        "database_name": "",
                        "description": DESCRIPTION,
                        "id": DATA_CONNECTOR_2_ID,
                        "name": DATABASE_NAME,
                        "options": S3_OPTIONS,
                        "secret_name": SECRET_NAME,
                        "type": "SOURCE",
                        "updated_at": "2023-01-18T00:18:01.000000",
                    },
                ],
            },
            headers={"Content-Type": "application/json"},
        )

        result = runner.invoke(
            list,
            env={"GANTRY_API_KEY": "test", "GANTRY_LOGS_LOCATION": UNITTEST_HOST},
        )

        assert result.exit_code == 0
        assert "--> A list of registered data connectors" in result.output
        assert DATA_CONNECTOR_1_ID in result.output
        assert DATA_CONNECTOR_2_ID in result.output


@pytest.mark.parametrize("connection_type", SUPPORTED_CONNECTION_TYPES)
def test_delete_data_connector(connection_type):
    """ """
    runner = CliRunner()
    cli_args = ["--name", DATA_CONNECTOR_NAME]

    with responses.RequestsMock() as resp:
        resp.add(
            resp.DELETE,
            f"{UNITTEST_HOST}/api/v1/data-connectors/sources/{DATA_CONNECTOR_NAME}",
            json={
                "response": "ok",
                "data": {
                    "connection_type": connection_type,
                    "created_at": "2023-01-18T00:18:01.000000",
                    "database_name": DATABASE_NAME,
                    "description": DESCRIPTION,
                    "id": DATA_CONNECTOR_1_ID,
                    "name": DATABASE_NAME,
                    "options": {"schema_name": "PUBLIC", "table_name": "GANTRY_EVENTS"},
                    "secret_name": SECRET_NAME,
                    "type": "SOURCE",
                    "updated_at": "2023-01-18T00:18:01.000000",
                },
            },
            headers={"Content-Type": "application/json"},
        )

        result = runner.invoke(
            delete,
            cli_args,
            env={"GANTRY_API_KEY": "test", "GANTRY_LOGS_LOCATION": UNITTEST_HOST},
        )

        assert result.exit_code == 0
        assert "--> Deleted the source data connector" in result.output
        assert DATA_CONNECTOR_1_ID in result.output


def test_list_pipelines_given_appname_status():
    """
    Tests that correct message displays when list command returns pipelines
    """
    runner = CliRunner()
    cli_args = ["--app-name", APP_NAME, "--status", DATA_CONNECTOR_ENABLED_STATUS_CMD]
    url = (
        f"{UNITTEST_HOST}/api/v1/data-connectors/pipelines?app_name={APP_NAME}"
        f"&enabled_status={DATA_CONNECTOR_ENABLED_STATUS}"
    )
    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            url,
            json={
                "response": "ok",
                "data": {
                    "consumer_id": PIPELINE_CONSUMER_ID,
                    "consumer_type": "GANTRY_EVENTLOGGER",
                    "created_at": "2023-02-11 01:04:57.293993",
                    "disabled_at": None,
                    "id": PIPELINE_ID,
                    "metadata_id": METADATA_ID,
                    "name": "2218762f-a684-480c-b418-d5adcdcb5dc6'",
                    "options": {
                        "sink_options": {
                            "join_key": "C",
                            "model_name": "c532db16-445f-419c-bab1-6082aa691bed",
                            "organization_id": "75c22adf-0494-4a6c-94f6-2e66aa3cd390",
                            "sink_path": "a:/b",
                            "version": "1.0.0",
                        },
                        "source_options": {
                            "feedbacks": [],
                            "global_tags": {"environment": "dev"},
                            "inputs": ["A"],
                            "join_key": "C",
                            "outputs": ["B"],
                            "row_level_tags": ["E"],
                            "timestamp": "D",
                        },
                    },
                    "type": "glueetl",
                    "updated_at": "2023-02-11 01:04:57.293993",
                },
            },
        )

        result = runner.invoke(
            list_pipelines,
            cli_args,
            env={
                "GANTRY_API_KEY": "test",
                "GANTRY_LOGS_LOCATION": UNITTEST_HOST,
                "GANTRY_DATASET_WORKING_DIR": "/tmp",
            },
        )
        assert result.exit_code == 0
        response = (
            f'--> The following pipelines are given for "{APP_NAME}" app with status'
            f' "{DATA_CONNECTOR_ENABLED_STATUS_CMD}"'
        )
        assert response in result.output

        assert (PIPELINE_ID) in result.output
        assert (PIPELINE_CONSUMER_ID) in result.output


def test_list_pipeline_ops_given_appname_status():
    """
    Tests that correct message displays when list command returns pipeline operations
    """
    runner = CliRunner()
    cli_args = [
        "--pipeline-name",
        PIPELINE_NAME,
        "--app-name",
        APP_NAME,
        "--status",
        PIPELINE_OPERATION_STATUS,
    ]

    url = (
        f"{UNITTEST_HOST}/api/v1/data-connectors/pipeline-operations?app_name={APP_NAME}"
        f"&pipeline_name={PIPELINE_NAME}&status={PIPELINE_OPERATION_STATUS}"
    )
    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            url,
            json={
                "response": "ok",
                "data": {
                    "consumer_operation_id": "5c327f88-a9a8-11ed-b6e2-6e131c5595a6",
                    "created_at": "Sat, 11 Feb 2023 01:06:48 GMT",
                    "id": PIPELINE_OPERATION_ID,
                    "metadata_id": METADATA_ID,
                    "options": None,
                    "pipeline_id": PIPELINE_ID,
                    "status": "QUEUED",
                    "trigger_id": None,
                    "trigger_type": "DIRECT",
                    "updated_at": "Sat, 11 Feb 2023 01:06:48 GMT",
                },
            },
        )

        result = runner.invoke(
            list_pipeline_operations,
            cli_args,
            env={
                "GANTRY_API_KEY": "test",
                "GANTRY_LOGS_LOCATION": UNITTEST_HOST,
                "GANTRY_DATASET_WORKING_DIR": "/tmp",
            },
        )
        assert result.exit_code == 0
        response = (
            f"--> The following pipeline operations are given for APP={APP_NAME} "
            f"and PIPELINE={PIPELINE_NAME} and STATUS={PIPELINE_OPERATION_STATUS}"
        )
        assert response in result.output
        assert METADATA_ID in result.output
        assert PIPELINE_OPERATION_ID in result.output
        assert PIPELINE_ID in result.output


def test_delete_pipeline():
    """Deletes a data connector pipeline given app_name and pipeline_name"""
    runner = CliRunner()
    cli_args = [
        "--pipeline-name",
        PIPELINE_NAME,
        "--app-name",
        APP_NAME,
        "--app-version",
        APP_VERSION,
    ]

    url = (
        f"{UNITTEST_HOST}/api/v1/data-connectors/pipelines?pipeline_name={PIPELINE_NAME}"
        f"&app_name={APP_NAME}&app_version={APP_VERSION}"
    )
    with responses.RequestsMock() as resp:
        resp.add(
            resp.DELETE,
            url,
            json={
                "response": "ok",
                "data": {
                    "id": PIPELINE_ID,
                    "created_at": "2023-02-16 01:01:01.999999",
                    "disabled_at": None,
                    "updated_at": "2023-02-16 01:01:01.999999",
                    "name": PIPELINE_NAME,
                    "options": {
                        "feedbacks": [],
                        "inputs": ["A"],
                        "outputs": ["B"],
                    },
                },
            },
            headers={"Content-Type": "application/json"},
        )

        result = runner.invoke(
            delete_pipeline,
            cli_args,
            env={"GANTRY_API_KEY": "test", "GANTRY_LOGS_LOCATION": UNITTEST_HOST},
        )

        assert result.exit_code == 0
        assert "--> Deleted the pipeline" in result.output
        assert PIPELINE_ID in result.output
        assert PIPELINE_NAME in result.output



import datetime
import json
import logging
import uuid
from decimal import Decimal

import dateutil
import numpy as np
import pandas as pd

_NUMPY_TYPE_MAP = {
    np.float32: float,
    np.float64: float,
    np.int32: int,
    np.int64: int,
    np.bool_: bool,
}

logger = logging.getLogger(__name__)


class EventEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, uuid.UUID):
            return str(obj)
        elif isinstance(obj, datetime.datetime):
            return datetime_isoformat(obj)
        elif type(obj) in _NUMPY_TYPE_MAP.keys():
            new_type = _NUMPY_TYPE_MAP[type(obj)]
            return new_type(obj)
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)


def datetime_isoformat(dt: datetime.datetime) -> str:
    iso_str = dt.isoformat()
    if dt.tzinfo is None or dt.tzinfo.utcoffset(dt) is None:
        # check if datetime is naive
        # https://docs.python.org/3/library/datetime.html#determining-if-an-object-is-aware-or-naive
        return f"{iso_str}Z"

    return iso_str


def serialize_datetime_like(value):
    error_msg = "Tried to serialize {} as datetime, but not in valid ISOformat".format(value)
    if isinstance(value, datetime.datetime):
        return datetime_isoformat(value)
    elif isinstance(value, str):
        try:
            dateutil.parser.parse(value)
            return value
        except ValueError:
            # Re-raise the exception
            raise ValueError(error_msg)

    raise ValueError(error_msg)


def serializable_value(value):  # noqa: C901
    if isinstance(value, uuid.UUID):
        return str(value)
    if isinstance(value, datetime.datetime):
        return datetime_isoformat(value)
    if isinstance(value, np.ndarray):
        return list_encoder(value.tolist())
    if isinstance(value, list) or isinstance(value, tuple):
        return list_encoder(value)
    if isinstance(value, dict):
        return dict_encoder(value)
    if isinstance(value, float) and (np.isnan(value) or np.isinf(value)):
        return None
    if isinstance(value, pd.Series):
        return list_encoder(value.to_list())
    if isinstance(value, pd.DataFrame):
        return list_encoder(value.to_dict("records"))
    if isinstance(value, Decimal):
        return float(value)
    if pd.isna(value):
        return None
    if type(value) in _NUMPY_TYPE_MAP.keys():
        new_type = _NUMPY_TYPE_MAP[type(value)]
        return new_type(value)

    if (
        not isinstance(value, str)
        and not isinstance(value, int)
        and not isinstance(value, bool)
        and not isinstance(value, float)
    ):
        logger.warn("%s of type %s is not serializable", value, type(value))

    return value


def list_encoder(li: list) -> list:
    serializable = list(li)
    for i in range(len(li)):
        value = li[i]
        serializable[i] = serializable_value(value)
    return serializable


def serializable_key(key):
    if key is None:
        return "None"
    if isinstance(key, uuid.UUID):
        return str(key)
    return key


def dict_encoder(d: dict) -> dict:
    serializable = {}
    for key, value in d.items():
        serializable[serializable_key(key)] = serializable_value(value)
    return serializable



import pytest

HOST = "https://test-api"
DATASET_NAME = "show_diff_dataset"
USER_EMAIL = "test@gantry.dev"
DATASET_ID = "77494dac-be4d-4c55-81bd-b16cf19f1297"
BUCKET_NAME = "test-bucket"
AWS_REGION = "us-west-2"
COMMIT_MSG = "test commit message"
MANIFEST_VERSION_ID = "mock_version_id"
S3_PREFIX = "org_id/datasets/v1"
CONF_OBJ_KEY = f"{S3_PREFIX}/{DATASET_NAME}/test-uuid-1/dataset_config.yaml"
README_OBJ_KEY = f"{S3_PREFIX}/{DATASET_NAME}/test-uuid-1/README.md"
IMG_OBJ_KEY = f"{S3_PREFIX}/{DATASET_NAME}/test-uuid-1/artifacts/kitten.png"
CSV_OBJ_KEY = f"{S3_PREFIX}/{DATASET_NAME}/test-uuid-1/tabular_manifests/feedback.csv"
MANIFEST_OBJ_KEY = f"{S3_PREFIX}/{DATASET_NAME}/.dataset_metadata/.gantry_manifest.jsonl"


@pytest.fixture(scope="function")
def test_dataset():
    return {
        "id": DATASET_ID,
        "name": DATASET_NAME,
        "organization_id": "org_id",
        "created_at": "Tue, 04 Oct 2022 22:15:25 GMT",
        "bucket_name": BUCKET_NAME,
        "aws_region": AWS_REGION,
        "s3_prefix": S3_PREFIX,
        "disabled": False,
    }


@pytest.fixture(scope="function")
def commit_history():
    return [
        {
            "created_at": "Wed, 05 Oct 2022 05:42:13 GMT",
            "created_by": "db459d6d-c83b-496d-b659-e48bca971156",
            "dataset_id": DATASET_ID,
            "id": "8ee0f6d5-c84c-473a-931a-5148b2e704d7",
            "is_latest_commit": True,
            "message": COMMIT_MSG,
            "metadata_s3_file_version": MANIFEST_VERSION_ID,
            "parent_commit": "eb2e7242-3340-4edf-8366-90d4fce897ce",
        },
        {
            "created_at": "Tue, 04 Oct 2022 22:15:25 GMT",
            "created_by": "db459d6d-c83b-496d-b659-e48bca971156",
            "dataset_id": DATASET_ID,
            "id": "eb2e7242-3340-4edf-8366-90d4fce897ce",
            "is_latest_commit": False,
            "message": "initial dataset commit",
            "metadata_s3_file_version": "PHHV3BI3euX_LbOiV5EjAouxHuX3ZBFh",
            "parent_commit": None,
        },
    ]



import atexit
import json
import logging
import os
import uuid
from dataclasses import dataclass
from queue import Queue
from typing import Callable, Dict, List, Optional, Union

import backoff
import requests

from gantry.const import PROD_API_URL
from gantry.exceptions import GantryRequestException

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore

from pydantic import BaseModel

from gantry.api_client import APIClient
from gantry.applications.core import Application
from gantry.applications.llm_utils import _get_prompt_inputs
from gantry.logger.consumer import BatchConsumer
from gantry.logger.stores import (
    BACKOFF_INTERVAL_FOR_ERROR_CODE_SECS,
    BACKOFF_INTERVAL_FOR_TIMEOUT_SECS,
    BATCH_UPLOAD_MAX_TRIES,
    BaseLogStore,
    backoff_giveup_handler,
    backoff_retry_handler,
    non_retry_error,
)

logger = logging.getLogger(__name__)


OPENAI_COMPLETION_MODELS = [
    "text-davinci-001",
    "text-davinci-002",
    "text-davinci-003",
    "text-curie-001",
    "text-babbage-001",
    "text-ada-001",
]


@dataclass(frozen=True)
class VersionDetails:
    """
    A dataclass representing the details of a version.

    Attributes:
        config: (dict): Dictionary of key and value pairs that specify the data stored in the
            version.
        description: (str): Description of the version.
        app_name: (str): Name of the application.
        version_id: (uuid.UUID): UUID of the version.
        version_number: (int): Version number of the version.
    """

    config: dict
    description: str
    app_name: str
    version_id: uuid.UUID
    version_number: int


class OpenAIParams(BaseModel):
    temperature: float = 0.7
    max_tokens: int = 256
    top_p: float = 1.0
    frequency_penalty: float = 0.0
    presence_penalty: float = 0.0


class CompletionAPILogStore(BaseLogStore):
    """Log store implementation for Completion.

    Only streaming logging is supported. This implementation
    uses the completion logging endpoint for ingestion and does
    not perform data validation.

    """

    def __init__(
        self,
        api_client,
        consumer_factory=BatchConsumer,
    ):
        self._api_client = api_client

        self.num_consumer_threads: int = 1
        self.queue: Queue = Queue()

        self.consumers = []
        self._consumer_factory = consumer_factory
        # On program exit, allow the consumer thread to exit cleanly.
        # This prevents exceptions and a messy shutdown when the
        # interpreter is destroyed before the daemon thread finishes
        # execution.
        atexit.register(self._join)
        for _ in range(self.num_consumer_threads):
            consumer = self._consumer_factory(self.queue, self.consumer_func)
            self.consumers.append(consumer)
            consumer.start()

    def log(self, application: str, event: dict) -> None:
        """
        Logs a prediction or feedback event.

        Args:
            application: Name of the application
            event: Data to logs as event body
        """
        self.queue.put(event)

    def log_batch(self, application: str, events: List[dict]) -> None:
        raise NotImplementedError("Completion logging does not support batch.")

    def consumer_func(self, batch):
        # Catch all errors and do not raise in order
        # for thread consumer to continue running.
        try:
            for event in batch:
                self.send_completion_event(json.loads(event))
        except Exception as e:
            logger.error("Internal error sending events: %s", e)

    @backoff.on_exception(
        backoff.constant,
        (requests.Timeout, requests.ConnectionError),
        interval=BACKOFF_INTERVAL_FOR_TIMEOUT_SECS,
        max_tries=BATCH_UPLOAD_MAX_TRIES,
        on_backoff=backoff_retry_handler,
        on_giveup=backoff_giveup_handler,
        jitter=None,
    )
    @backoff.on_exception(
        backoff.constant,
        GantryRequestException,
        max_tries=BATCH_UPLOAD_MAX_TRIES,
        interval=BACKOFF_INTERVAL_FOR_ERROR_CODE_SECS,
        on_backoff=backoff_retry_handler,
        on_giveup=backoff_giveup_handler,
        giveup=non_retry_error,
        jitter=None,
    )
    def send_completion_event(self, event: Dict) -> None:
        response = self._api_client.request(
            "POST",
            "/api/v1/logging/completions",
            json=event,
            raise_for_status=True,
        )
        if response.get("response") != "ok":
            logger.error("Failed to log events. Response = %s", response)

    def _join(self):
        for consumer in self.consumers:
            consumer.pause()
            try:
                consumer.join()
            except RuntimeError:
                # consumer thread has not started
                pass


class CompletionApplication(Application):
    """
    A class representing a completion application, a subclass of Application. This has
    special logging methods for logging completions.
    """

    def __init__(
        self,
        name: str,
        api_client: APIClient,
        id: Optional[uuid.UUID] = None,
        organization_id: Optional[uuid.UUID] = None,
        log_store_factory: Callable[[APIClient], BaseLogStore] = CompletionAPILogStore,
    ):
        super().__init__(name, api_client, id, organization_id)
        self._configuration_name = f"default_{name}_configuration"
        # TODO: validate name

        self._log_store = log_store_factory(api_client)

    def create_version(
        self,
        prompt_template: str,
        description: str,
        model: str,
        model_params: Optional[dict] = None,
        prompt_inputs: Optional[List[str]] = None,
    ) -> VersionDetails:
        """
        Creates a new version of a prompt template associated to this application.

        Example usage:

        .. code-block:: python

            import gantry

            my_llm_app = gantry.get_application("my_llm_app")

            my_llm_app.create_version(
                prompt_template="This is a prompt template. {{input1}} {{input2}}",
                description="My first version",
                model="text-davinci-001",
                model_params={"temperature": 0.5},
                prompt_inputs=["input1", "input2"],
            )

        Example output:

        .. code-block:: python

            VersionDetails(
                config={
                    "model": "text-davinci-002",
                    "params": {
                        "frequency_penalty": 0.0,
                        "max_tokens": 16,
                        "presence_penalty": 0.0,
                        "temperature": 0.5,
                        "top_p": 1.0
                    },
                    "prompt": "This is a prompt template. {{input1}} {{input2}}",
                    "prompt_inputs": ["input1", "input2"],
                },
                description="My first version",
                app_name="my_llm_app",
                version_id=UUID("a1b2c3d4-e5f6-4a3b-8c2d-1e2f3a4b5c6d"),
                version_number=1,
            )



        Args:
            prompt_template (str): The prompt template to use for this version. Denote variables
                like this: `{{variable}}`
            description (str): A description of this version.
            model (str): The OpenAI model to use for this version. Must be a valid OpenAI
                completion text model.
            model_params (dict, optional): A dictionary of model parameters to use for this prompt.
                Valid keys are: temperature, max_tokens, top_p, frequency_penalty, presence_penalty.
                See OpenAI's documentation for more details.
            prompt_inputs (List[str], optional): A list of input names to use for this prompt,
                when using prompt variables. If not provided, the function will attempt to extract
                the input names from the prompt template.

        Returns:
            VersionDetails: A dataclass containing the configuration data, description, application
            name, version ID, and version number of the newly created version.
        """
        if model not in OPENAI_COMPLETION_MODELS:
            raise ValueError(f"Invalid model '{model}'. Must be one of {OPENAI_COMPLETION_MODELS}")

        model_params = model_params or {}
        params = OpenAIParams(**model_params).dict()

        configuration_data = {
            "prompt": prompt_template,
            "prompt_inputs": prompt_inputs or _get_prompt_inputs(prompt_template),
            "model": model,
            "params": params,
        }

        res = self._api_client.request(
            "POST",
            f"/api/v1/cms/{self._name}/configurations/{self._configuration_name}/versions",
            json={
                "configuration_data": configuration_data,
                "description": description,
            },
            raise_for_status=True,
        )
        logger.info(
            "Version %s created for application '%s'",
            res["data"]["version"],
            self._name,
        )
        host = os.environ.get("GANTRY_LOGS_LOCATION", PROD_API_URL)
        link_to_versions = "{}/applications/{}/versions/".format(host, self._name)
        logger.info("See the new version here: %s", link_to_versions)

        return VersionDetails(
            config=configuration_data,
            description=description,
            app_name=self._name,
            version_id=uuid.UUID(res["data"]["id"]),
            version_number=res["data"]["version"],
        )

    def get_version(self, version: Literal["prod", "test", "latest"]) -> Optional[VersionDetails]:
        """
        Returns the prompt configuration data for a given version of this application.

        Example usage:

        .. code-block:: python

            import gantry

            my_llm_app = gantry.get_application("my_llm_app")

            my_llm_app.get_version("prod")

        Example output:

        .. code-block:: python

            VersionDetails(
                config={
                    "model": "text-davinci-002",
                    "params": {
                        "frequency_penalty": 0.0,
                        "max_tokens": 16,
                        "presence_penalty": 0.0,
                        "temperature": 0.5,
                        "top_p": 1.0
                    },
                    "prompt": "This is a prompt template. {{input1}} {{input2}}",
                    "prompt_inputs": ["input1", "input2"],
                },
                description="My first version",
                app_name="my_llm_app",
                version_id=UUID("a1b2c3d4-e5f6-4a3b-8c2d-1e2f3a4b5c6d"),
                version_number=1,
            )

        Args:
            version (Literal["prod", "test", "latest"]): The version to get. Can be one of "prod",
                "test", or "latest". When "latest" is used, the latest version will be returned,
                regardless of deployment status. When "prod" or "test" is used, the latest version
                that has been deployed to that environment will be returned.

        Returns:
            Optional[VersionDetails]: A dataclass containing the configuration data, description,
            application name, version ID, and version number of the requested version. If no
            version is found, returns None.
        """
        version_data: dict

        try:
            if version == "latest":
                # for latest version, we get a list of all versions and pick the most recent
                res = self._api_client.request(
                    "GET",
                    f"/api/v1/cms/{self._name}/configurations/{self._configuration_name}/versions",
                    raise_for_status=True,
                )
                version_data = res["data"][0]
            else:
                # for releases, we can get the latest release by environment
                res = self._api_client.request(
                    "GET",
                    f"/api/v1/cms/{self._name}/configurations/{self._configuration_name}/releases",
                    params={"latest": True, "env": version},
                    raise_for_status=True,
                )
                version_data = res["data"][0]["version"]
        except IndexError:
            logger.warn("No version found for '%s'", version)
            return None

        # now that we have the version, we can get the config data
        res = self._api_client.request(
            "GET",
            (
                f"/api/v1/cms/{self._name}/configurations/{self._configuration_name}"
                f"/versions/{version_data['id']}/data"
            ),
            raise_for_status=True,
        )
        config_data = res["data"][0]

        return VersionDetails(
            config=config_data,
            description=version_data["notes"],
            app_name=self._name,
            version_id=uuid.UUID(version_data["id"]),
            version_number=version_data["version"],
        )

    def log_completion(
        self,
        open_ai_api_request: Dict,
        open_ai_api_response: Dict,
        request_attributes: Optional[Dict] = None,
        response_attributes: Optional[Dict] = None,
        feedback: Optional[Dict] = None,
        selected_choice_index: int = 0,
        session_id: Optional[Union[str, uuid.UUID]] = None,
        tags: Optional[Dict] = None,
        version: Optional[int] = None,
    ):
        """
        Log request/response OpenAI pair into Gantry.

        Example usage:

        .. code-block:: python

            import gantry
            from gantry.applications.llm_utils import fill_prompt
            import openai
            from openai.util import convert_to_dict

            gantry.init()
            my_llm_app = gantry.get_application("my-llm-app")

            version = my_llm_app.get_version("test")
            config = version.config
            prompt = config['prompt']

            def generate(values):
                filled_in_prompt = fill_prompt(prompt, values)
                request = {
                    "model": "text-davinci-002",
                    "prompt": filled_in_prompt,
                }
                results = openai.Completion.create(**request)

                my_llm_app.log_completion(
                    open_ai_api_request=request,
                    open_ai_api_response=convert_to_dict(results),
                    request_attributes={"prompt_values": values},
                    version=version.version_number,
                )

                return results

        For more details on request/response content, see `OpenAI API Reference for Completion
        <https://platform.openai.com/docs/api-reference/completions/create>`_.

        Args:
            open_ai_api_request (dict): The OpenAI request.
                For more details on content, see `OpenAI API Reference for Completion
                <https://platform.openai.com/docs/api-reference/completions/create>`_.
            open_ai_api_response (dict): The OpenAI response.
                For more details on content, see `OpenAI API Reference for Completion
                <https://platform.openai.com/docs/api-reference/completions/create>`_.
            request_attributes (Optional[dict]): Additional inputs
            response_attributes (Optional[dict]): Additional outputs
            feedback (Optional[dict]): Optional feedback data. See
                https://docs.gantry.io/docs/logging-feedback-actuals for more information
                about logging feedback in Gantry.
            selected_choice_index (Optional[int]): The selected choice index. Defaults to 1.
            session_id (Optional[str or uuid.UUID]): Optional session
                to group data together.
            tags (Optional[dict]): Optional tags to add to the record.
            version (Optional[int]): The version number used.
        """
        tags = tags or {}
        if version:
            tags["version_number"] = version

        event = {
            "application": self._name,
            "open_ai_api_request": open_ai_api_request,
            "open_ai_api_response": open_ai_api_response,
            "request_attributes": request_attributes or {},
            "response_attributes": response_attributes or {},
            "feedback": feedback or {},
            "selected_choice_index": selected_choice_index,
            "session_id": session_id,
            "tags": tags,
        }

        self._log_store.log(self._name, event)

    def deploy_version(self, version_id: uuid.UUID, env: Literal["prod", "test"]) -> None:
        """
        Deploys a version in a specific environment.

        Example usage:

        .. code-block:: python

            import gantry

            my_llm_app = gantry.get_application("my_llm_app")
            version = my_llm_app.get_version("latest")
            my_llm_app.deploy_version(version.version_id, "prod")

        Args:
            version_id (uuid): The version id.
            env (str, 'prod' or 'test'): The target environment, can be 'prod' or 'test'.

        """
        _ = self._api_client.request(
            "PATCH",
            (f"/api/v1/cms/{self._name}/configurations/{self._configuration_name}" "/releases"),
            json={
                "version_id": str(version_id),
                "env": env,
                "description": "Deployment from SDK",
            },
            raise_for_status=True,
        )
        logger.info(f"Version {version_id} was successfully deployed in {env}")



from datetime import datetime, timedelta
import pandas as pd
import numpy as np
import uuid
import random
import logging

logger = logging.getLogger(__name__)

RANDOM_SEED = 43
DEFAULT_SAMPLE_SIZE = 2000
MIN_SAMPLE_SIZE = 200
MAX_SAMPLE_SIZE = 4000

# Generate random description
PRONOUNS = ["He", "She"]
VERBS = ["was", "is", "used to be", "wants to become"]
NOUNS = [
    "a doctor.",
    "an engineer.",
    "a infuluencer.",
    "a dancer.",
    "a lyft driver.",
    "a pilot" "a singer.",
    "a freelancer.",
    "a teacher.",
    "a professor.",
]


def _generate_batch_data(
    batch_size: int = 1000,
    cred_hist_prop: float = 0.7,
    pred_true_rate: float = 0.6,
    loan_repaid_rate: float = 0.8,
    random_seed: int = RANDOM_SEED,
) -> pd.DataFrame:
    """
    Generate a single batch data. You can control the data
    distribution by changing the input parameters

    Args:
        batch_size(int, defaults to 1000): size of the data
        cred_hist_prop(float, defaults to 0.7): Ratio of customers who have good credit history
        pred_true_rate(float, defaults to 0.6): Positive prediction ratio
        loan_repaid_rate(float, defaults to 0.8): Positive feedback ratio
        ramdom_seed(int, default to 43): random seed
    Returns:
        pd.DataFrame which contain the generated data
    """
    if any(param < 0 or param > 1 for param in [cred_hist_prop, pred_true_rate, loan_repaid_rate]):
        raise ValueError("Must be a value between 0 and 1")

    rng = np.random.default_rng(random_seed)
    credit_history_categorical = rng.choice(
        ["Good", "Bad"], batch_size, p=[cred_hist_prop, 1 - cred_hist_prop]
    )

    income = rng.normal(loc=50000, scale=15000, size=batch_size)
    income = np.clip(income, 10000, 200000).astype(int)

    predictions = rng.uniform(size=batch_size) < pred_true_rate
    loan_repaid = rng.uniform(size=batch_size) < loan_repaid_rate

    rand_pronouns = rng.choice(PRONOUNS, batch_size)
    rand_verbs = rng.choice(VERBS, batch_size)
    rand_nouns = rng.choice(NOUNS, batch_size)
    text = [" ".join([p, v, n]) for p, v, n in zip(rand_pronouns, rand_verbs, rand_nouns)]

    random.seed(random_seed)
    loan_ids = [str(uuid.UUID(int=random.getrandbits(128))) for _ in range(batch_size)]

    df = pd.DataFrame(
        {
            "loan_id": loan_ids,
            "credit_history": credit_history_categorical,
            "income": income,
            "description": text,
            "loan_repaid": loan_repaid,
            "prediction": predictions,
        }
    )
    df.credit_history = df.credit_history.astype("category")
    return df


def generate_demo_data(
    sample_size: int = DEFAULT_SAMPLE_SIZE, random_seed: int = RANDOM_SEED
) -> pd.DataFrame:
    """
    Function to generate synthetic data to be used for exploring Gantry features.
    This function will generate data points for the last 48 hours. The first 24
    hours' data has a different distribution with the second 24 hours' data.
    You can create views and compare the data drift/different.

    Note: data is randomly generated and should only be used for demo purposes.

    Args:
        sample_size (int, defaults to 2000): sample dataset size, sample size should
        be larger than 200 and smaller than 4000
        random_seed (int, default to 43): random seed
    Returns:
        pd.DataFrame which contain the generated data
    """
    if sample_size < MIN_SAMPLE_SIZE:
        sample_size = MIN_SAMPLE_SIZE
        logger.warn(
            f"sample_size smaller than {MIN_SAMPLE_SIZE}, will use {MIN_SAMPLE_SIZE} instead."
        )

    if sample_size > MAX_SAMPLE_SIZE:
        sample_size = MAX_SAMPLE_SIZE
        logger.warn(
            f"sample_size larger than {MAX_SAMPLE_SIZE}, will use {MAX_SAMPLE_SIZE} instead."
        )

    batch1_size = sample_size // 2 + sample_size % 2
    batch2_size = sample_size // 2

    batch_1 = _generate_batch_data(
        batch_size=batch1_size,
        cred_hist_prop=0.9,
        pred_true_rate=0.8,
        loan_repaid_rate=0.8,
        random_seed=random_seed,
    )

    batch_2 = _generate_batch_data(
        batch_size=batch2_size,
        cred_hist_prop=0.2,
        pred_true_rate=0.5,
        loan_repaid_rate=0.4,
        random_seed=random_seed,
    )

    end_time = datetime.now()
    mid_time = end_time - timedelta(seconds=86400)
    start_time = end_time - timedelta(seconds=172800)

    random.seed(random_seed)
    batch_1["timestamp"] = [
        start_time + timedelta(seconds=random.randint(0, 86399)) for _ in range(batch1_size)
    ]
    batch_2["timestamp"] = [
        mid_time + timedelta(seconds=random.randint(0, 86399)) for _ in range(batch2_size)
    ]

    return pd.concat([batch_1, batch_2])



from queue import Queue

import mock
import pytest
from freezegun import freeze_time

from gantry.logger import consumer

from ..conftest import TestingBatchIter, TestingBatchIterWithEmptyFlag


@pytest.fixture(scope="function")
def q():
    return Queue()


@pytest.fixture(scope="function")
def batch_consumer_factory(q):
    def wrapper(f):
        return consumer.BatchConsumer(queue=q, func=f, batch_iter=TestingBatchIter)

    return wrapper


def test_batch_iter_builder_size_limit(q):
    """Check that batch iterator caches elements
    until reaching batch_size_record_limit. Then
    returns cache in next consumed batch.
    """
    batch_iter = consumer._BatchIterBuilder(
        queue=q,
        batch_size_record_limit=50,
        single_event_limit_bytes=1000,
        batch_size_limit_bytes=1000,
        timeout_secs=0.1,
    )

    for i in range(100):
        q.put(i)

    it = iter(batch_iter)

    # First 50 items are empty until batch size gets to 50
    for i in range(50):
        assert next(it) == []

    # Batch with [b'0', b'1', ..., b'49']
    assert next(it) == [str(x).encode() for x in range(50)]

    for i in range(49):
        assert next(it) == []

    # Batch with [b'50', b'51', ..., b'99']
    assert next(it) == [str(x).encode() for x in range(50, 100)]

    # No more items in queue
    assert q.empty()


def test_batch_iter_builder_single_event_limit(q):
    """Check that batch iterator drops big events."""
    batch_iter = consumer._BatchIterBuilder(
        queue=q,
        batch_size_record_limit=1,
        single_event_limit_bytes=3,
        batch_size_limit_bytes=1000,
        timeout_secs=0.1,
    )

    q.put("1")
    q.put("333")  # This item will be ignored
    q.put("1")

    it = iter(batch_iter)

    assert next(it) == []
    assert next(it) == []
    assert next(it) == [b'"1"']
    assert next(it) == [b'"1"']

    # No more items in queue
    assert q.empty()


def test_batch_iter_builder_batch_size_limit_bytes(q):
    """Check that batch iterator flushed when cache gets big"""
    batch_iter = consumer._BatchIterBuilder(
        queue=q,
        batch_size_record_limit=100,
        single_event_limit_bytes=100,
        batch_size_limit_bytes=10,
        timeout_secs=0.1,
    )

    q.put("11")
    q.put("22")
    q.put("11")

    it = iter(batch_iter)

    assert next(it) == []
    assert next(it) == []
    assert next(it) == [b'"11"', b'"22"']
    assert next(it) == [b'"11"']

    # No more items in queue
    assert q.empty()


def test_batch_iter_builder_timeout(q):
    """Check that batch iterator timeouts while consuming on an empty queue"""
    batch_iter = consumer._BatchIterBuilder(
        queue=q,
        batch_size_record_limit=100,
        single_event_limit_bytes=100,
        batch_size_limit_bytes=10,
        timeout_secs=0.01,
    )

    it = iter(batch_iter)

    assert next(it) == []
    assert next(it) == []

    # No more items in queue
    assert q.empty()


def test_batch_iter_builder_flushing_after(q):
    """Check that batch iterator flushes after cache timeout"""
    batch_iter = consumer._BatchIterBuilder(
        queue=q,
        batch_size_record_limit=100,
        single_event_limit_bytes=100,
        batch_size_limit_bytes=100,
        timeout_secs=0.01,
        flush_after_secs=-1,
    )

    it = iter(batch_iter)

    q.put("1")
    q.put("2")

    assert next(it) == []
    assert next(it) == [b'"1"']
    assert next(it) == [b'"2"']


@freeze_time("2012-01-14 03:21:34")
def test_batch_iter_flush_and_reset(q):
    """Check that this method flushed, marks
    the task as done and reset the _last_flushed ckpt.
    """
    batch_iter = consumer._BatchIterBuilder(
        queue=q,
        batch_size_record_limit=100,
        single_event_limit_bytes=100,
        batch_size_limit_bytes=10,
        timeout_secs=0.1,
    )
    q.task_done = mock.Mock()

    batch_iter.current_json_dumped_batch = ["foobar"]
    ret = batch_iter._flush_and_reset("barbaz")

    q.task_done.assert_called_once()
    assert ret == ["foobar"]
    assert batch_iter.current_json_dumped_batch == ["barbaz"]
    assert batch_iter.current_json_dumped_batch_size_bytes == 6
    assert batch_iter._last_flushed == 1326511294


def test_batch_iter_empty(q):
    """Batch iterator is empty if the queue
    is empty and there are no events in the cache
    """
    batch_iter = consumer._BatchIterBuilder(
        queue=q,
        batch_size_record_limit=100,
        single_event_limit_bytes=100,
        batch_size_limit_bytes=10,
        timeout_secs=0.1,
    )

    assert batch_iter.empty()

    q.put("1")
    assert not batch_iter.empty()

    it = iter(batch_iter)

    assert next(it) == []
    assert not batch_iter.empty()

    assert next(it) == [b'"1"']
    assert batch_iter.empty()


def test_batch_consumer_consume(batch_consumer_factory, q):
    """Check that the consume method uses the provided function"""
    batch = []

    def func(event):
        batch.append(event)

    for i in range(100):
        q.put(i)

    batch_consumer = batch_consumer_factory(func)
    batch_iter = iter(batch_consumer._batch_iter(q))

    for i in range(100):
        batch_consumer.consume(batch_iter)

    assert batch == [[x] for x in range(100)]


def test_batch_consumer_pause(batch_consumer_factory, q):
    batch_consumer = batch_consumer_factory(None)
    assert batch_consumer.running

    batch_consumer.pause()

    assert not batch_consumer.running


def test_batch_consumer_run(batch_consumer_factory, q):
    """Consumer 'run' method should stop executing
    after consumer is paused and the iterator reports
    emptyness.
    """
    batch_iter = TestingBatchIterWithEmptyFlag(q)
    batch_iter._empty = False

    q.put("1")
    q.put("2")
    q.put("3")

    q.put("4")
    q.put("5")

    batches = []

    def func(event):
        batches.append(event)
        # After consuming '1', '2', '3'
        # the batch iterator will report
        # it is empty, so '4' and '5'
        # won't be consumed.
        if len(batches) == 3:
            batch_iter._empty = True

    def batch_iter_factory(q):
        return batch_iter

    batch_consumer = consumer.BatchConsumer(queue=q, func=func, batch_iter=batch_iter_factory)
    batch_consumer.pause()
    batch_consumer.run()

    assert not batch_consumer.running
    assert batches == [[b'"1"'], [b'"2"'], [b'"3"']]



from typing import List, Optional

from gantry.exceptions import GantryLoggingDataTypeError

MAX_CALL_DEPTH = 10


SUPPORTED_DRUID_TYPES = (int, str, float, bool, type(None))


def validate_logged_field_datatype(field, paths: Optional[List[str]] = None) -> None:
    """
    Validates that fields is:
        a valid primitive type
        OR
        a list that is no larger than 2 dimensions and contains valid primitive types
        OR
        a dict that is no larger than 10 levels and contains valid primitive types
    """
    paths = paths or []

    if isinstance(field, list) or isinstance(field, tuple):
        for i, item in enumerate(field):
            # a list item can be a primitive type
            if _validate_primitive_type(item):
                continue

            # or it can be a 1-dimensional list containing primitive types
            if isinstance(item, list):
                for j, nested_item in enumerate(item):
                    if not _validate_primitive_type(nested_item):
                        full_path = ".".join(paths) + f"[{i}][{j}]"
                        field_type = type(field)
                        raise GantryLoggingDataTypeError(
                            (
                                "{} is not a valid Gantry datatype. "
                                "Lists must be 1- or 2-dimensional and contain only primitive "
                                "types."
                            ).format(
                                full_path,
                            )
                        )
            else:
                full_path = ".".join(paths) + f"[{i}]"
                item_type = type(item)
                raise GantryLoggingDataTypeError(
                    (
                        "{} with type {} is not a valid Gantry datatype. "
                        "Lists must be 1- or 2-dimensional and contain only primitive types."
                    ).format(
                        full_path,
                        item_type,
                    )
                )
    elif isinstance(field, dict):
        for k, v in field.items():
            validate_logged_field_datatype(k, paths=(paths + [str(k), "key"]))
            validate_logged_field_datatype(v, paths=(paths + [str(k), "value"]))
    else:
        if _validate_primitive_type(field):
            return

        full_path = ".".join(paths)
        field_type = type(field)
        raise GantryLoggingDataTypeError(
            "{} with type {} is not a valid Gantry datatype".format(full_path, field_type)
        )


def _validate_primitive_type(field) -> bool:
    return any(isinstance(field, t) for t in SUPPORTED_DRUID_TYPES)



from collections import OrderedDict

import mock
import pandas as pd
import pytest
import responses
from pandas.testing import assert_frame_equal
from responses import matchers

from gantry.query.core.dataframe import Filters, GantryDataFrame

from ..conftest import DATA, END_TIME, MULTIPLE_FEATURES_INT_DATA, ORIGIN, START_TIME


@pytest.fixture
@mock.patch("gantry.query.core.dataframe.GantryDataFrame._build_series")
def df(mock_populate, api_client_obj):
    with responses.RequestsMock() as rsps:
        rsps.add(
            responses.GET,
            "{}/api/v1/models/foobar/schemas".format(ORIGIN),
            json={"response": "ok", "data": DATA},
        )

        return GantryDataFrame(
            api_client=api_client_obj,
            application="foobar",
            version="1.2.3",
            env="dev",
            start_time=START_TIME,
            end_time=END_TIME,
            filters=[],
            tags={"bar": "baz"},
        )


@pytest.fixture
@mock.patch(
    "gantry.query.core.queryframe.GantryQueryFrame.metadata",
    return_value=MULTIPLE_FEATURES_INT_DATA,
    new_callable=mock.PropertyMock,
)
@mock.patch("gantry.query.core.dataframe.get_application_node_id")
def df_with_int(get_application_node_id_mock, metadata, api_client_obj):
    get_application_node_id_mock.return_value = "12345"
    return GantryDataFrame(
        api_client=api_client_obj,
        application="foobar",
        version="1.2.3",
        env="dev",
        start_time=START_TIME,
        end_time=END_TIME,
        filters=[],
    )


def test_build_series(df):
    content = {
        "end_time": "2008-09-10 23:58:23",
        "start_time": "2008-09-03 20:56:35",
        "version": "1.2.3",
    }

    with responses.RequestsMock() as rsps:
        rsps.add(
            responses.GET,
            "{}/api/v1/models/foobar/schemas".format(ORIGIN),
            json={"response": "ok", "data": DATA},
            match=[matchers.query_param_matcher(content)],
        )

        series = df._build_series()
        assert len(series) == 3

        assert "pred_datanode" in series
        assert series["pred_datanode"].dtype == "intA"
        assert series["pred_datanode"].series_type == "A"

        assert "feed_datanode" in series
        assert series["feed_datanode"].dtype == "intB"
        assert series["feed_datanode"].series_type == "B"

        assert "proj_datanode" in series
        assert series["proj_datanode"].dtype == "intC"
        assert series["proj_datanode"].series_type == "C"

        try:
            repr(df)
            repr(series)
        except Exception as e:
            assert False, f"__repr__ failed due to: {e}"


@mock.patch("gantry.query.main.create_view")
def test_create_view(mock_create_view, df):
    df.filters = [{"this-is": "a-filter"}]
    df.create_view("my-view-name")
    mock_create_view.assert_called_once_with(
        name="my-view-name",
        application="foobar",
        start_time=START_TIME,
        end_time=END_TIME,
        version="1.2.3",
        tag_filters={"bar": "baz"},
        data_filters=[{"this-is": "a-filter"}],
    )


def test_clone_with_extra_filters_error(df, series_obj):
    f = Filters._from_single_filter({"foo": "bar"}, series_obj)
    with pytest.raises(ValueError):
        df._clone_with_extra_filters(f)


@mock.patch(
    "gantry.query.core.queryframe.GantryQueryFrame.metadata",
    return_value=MULTIPLE_FEATURES_INT_DATA,
    new_callable=mock.PropertyMock,
)
@mock.patch("gantry.query.core.dataframe.get_application_node_id")
def test_clone_with_extra_filters(
    get_application_node_id_mock, metadata, df_factory, series_obj, api_client_obj
):
    get_application_node_id_mock.return_value = "12345"
    df = GantryDataFrame(
        api_client=api_client_obj,
        application="foobar",
        version="1.2.3",
        env="dev",
        start_time=START_TIME,
        end_time=END_TIME,
        filters=[{"bar": "baz"}],
        tags={"foo": "bar"},
    )
    series_obj.parent_dataframe = df
    f = Filters._from_single_filter({"foo": "bar"}, series_obj)

    another_df = df._clone_with_extra_filters(f)

    assert another_df.filters == [{"bar": "baz"}, {"foo": "bar"}]
    assert another_df.tags == {"foo": "bar"}


@mock.patch(
    "gantry.query.core.queryframe.GantryQueryFrame.metadata",
    return_value=DATA,
    new_callable=mock.PropertyMock,
)
def test_dtypes(meta, df):
    assert isinstance(df.dtypes, pd.Series)
    expected = pd.Series(
        {
            "feed_datanode": "intB",
            "pred_datanode": "intA",
            "proj_datanode": "intC",
        }
    )
    pd.testing.assert_series_equal(df.dtypes, expected)


def test_inexistent_attr(df):
    with pytest.raises(AttributeError, match="some_inexistent_attr is not a valid attribute"):
        _ = df.some_inexistent_attr


def test_inexistent_key(df):
    with pytest.raises(KeyError, match="some_inexistent_col is not a valid column"):
        _ = df["some_inexistent_col"]


@pytest.mark.parametrize("n", [1, 5, 10, 11])
@mock.patch("gantry.query.core.dataframe.GantryQueryFrame._populate_raw_request")
@mock.patch("gantry.query.core.dataframe.GantryQueryFrame._prepare_pandas_dataframe")
def test_head(mock_df, mock_data, n, df):
    total_num_events = 10
    all_events = list(range(total_num_events))
    mock_data.return_value = all_events[:n]
    mock_df.return_value = pd.DataFrame({"foo": mock_data.return_value})
    pd.testing.assert_frame_equal(df.head(n), mock_df.return_value)
    mock_data.assert_called_with(query_filters=[], ascending=True, n=n, tags={"bar": "baz"})


@pytest.mark.parametrize("n", [1, 5, 10, 11])
@mock.patch("gantry.query.core.dataframe.GantryQueryFrame._populate_raw_request")
@mock.patch("gantry.query.core.dataframe.GantryQueryFrame._prepare_pandas_dataframe")
def test_tail(mock_df, mock_data, n, df):
    total_num_events = 10
    all_events = list(range(total_num_events))
    mock_data.return_value = all_events[:n]
    mock_df.return_value = pd.DataFrame({"foo": mock_data.return_value})
    pd.testing.assert_frame_equal(df.tail(n), mock_df.return_value)
    mock_data.assert_called_with(query_filters=[], ascending=False, n=n, tags={"bar": "baz"})


@mock.patch("gantry.query.core.dataframe.GantryQueryFrame._populate_raw_request")
@mock.patch("gantry.query.core.dataframe.GantryQueryFrame._prepare_pandas_dataframe")
def test_fetch(mock_df, mock_data, df):
    total_num_events = 10
    all_events = list(range(total_num_events))
    mock_data.return_value = all_events
    mock_df.return_value = pd.DataFrame({"foo": mock_data.return_value})
    pd.testing.assert_frame_equal(df.fetch(), mock_df.return_value)
    mock_data.assert_called_with(query_filters=[], ascending=True, n=None, tags={"bar": "baz"})


@mock.patch("gantry.query.core.dataframe.GantrySeries._describe")
@mock.patch("gantry.query.core.dataframe.GantryDataFrame._build_series")
def test_describe(mock_build_series, mock_series_describe, df, series_obj_1):
    df.series = OrderedDict({"s1": series_obj_1, "s2": series_obj_1})
    mock_series_describe.side_effect = [{"foo": "bar"}, {"foo": "baz"}]
    expected = pd.DataFrame(
        [
            ["foobar", "foobar"],
            [START_TIME, START_TIME],
            [END_TIME, END_TIME],
            ["bar", "baz"],
        ],
        index=["application", "from", "to", "foo"],
        columns=["s1", "s2"],
    )
    assert_frame_equal(df.describe(), expected)


def test_parent_filters(df_factory, series_obj_factory):
    # Create df with a series feature and existing filters
    existing_filters = [mock.Mock()]
    df = df_factory(existing_filters)
    series = series_obj_factory(datatype="int", parent_dataframe=df)
    df.series = {"series_name": series}

    # This action shouldn't change the df filters
    _ = df["series_name"] > 10
    assert len(df.filters) == 1


@mock.patch(
    "gantry.query.core.queryframe.GantryQueryFrame.metadata",
    return_value=MULTIPLE_FEATURES_INT_DATA,
    new_callable=mock.PropertyMock,
)
@mock.patch("gantry.query.core.dataframe.get_application_node_id")
def test_parent_filters_resolve(get_application_node_id_mock, metadata, api_client_obj):
    get_application_node_id_mock.return_value = "12345"
    df = GantryDataFrame(
        api_client=api_client_obj,
        application="foobar",
        version="1.2.3",
        env="dev",
        start_time=START_TIME,
        end_time=END_TIME,
        filters=[],
    )

    _ = df[df["A"] > 10]
    other_df_with_filters = df[df["A"] < 20]

    # This series should have only 1 filter, the < 20 restriction
    # If this doesn't happen, it means that all mehtods using
    # _aggregate_query will not work.
    assert len(other_df_with_filters["A"].parent_dataframe.filters) == 1


def test_filtering_using_other_df(df, df_with_int):
    with pytest.raises(ValueError):
        _ = df[df_with_int["A"] > 10]


VIEWS_DATA = [
    {
        "name": "foobar",
        "start_time": START_TIME.isoformat(),
        "end_time": END_TIME.isoformat(),
        "tag_filters": {"env": "barbaz"},
        "data_filters": [{"XYZ": "123"}],
    },
]


@mock.patch(
    "gantry.query.core.queryframe.GantryQueryFrame.metadata",
    return_value=MULTIPLE_FEATURES_INT_DATA,
    new_callable=mock.PropertyMock,
)
@mock.patch("gantry.query.core.dataframe.get_application_node_id")
@mock.patch("gantry.query.core.dataframe.get_start_end_time_from_view")
def test_from_view_factory(mock_get_times, get_application_node_id_mock, metadata, api_client_obj):
    mock_get_times.return_value = (START_TIME, END_TIME)
    get_application_node_id_mock.return_value = "12345"
    version = "0.1.2"

    with responses.RequestsMock() as rsps:
        rsps.add(
            responses.GET,
            "{}/api/v1/applications/app/views?version={}".format(ORIGIN, version),
            json={"response": "ok", "data": VIEWS_DATA},
        )

        rsps.add(
            responses.GET,
            "{}/api/v1/models/app/schemas".format(ORIGIN),
            json={"response": "ok", "data": {"version": version}},
        )

        df = GantryDataFrame.from_view(
            api_client=api_client_obj,
            application="app",
            view="foobar",
            version=None,
            environment="barbaz",
            tags={"foo": "bar"},
        )

    assert df.filters == [{"XYZ": "123"}]
    assert df.query_info.start_time == START_TIME
    assert df.query_info.end_time == END_TIME
    assert df.query_info.version == "0.1.2"
    assert df.query_info.environment == "barbaz"
    assert df.query_info.application == "app"
    assert df.query_info.application_node_id == "12345"
    assert df.api_client is api_client_obj
    assert df.series.keys() == set(["A", "B"])

    mock_get_times.assert_called_with(VIEWS_DATA[0])


@pytest.mark.parametrize(
    ["view", "environment"],
    [
        ("invalid", "barbaz"),
        ("foobar", "invalid"),
    ],
)
@mock.patch("gantry.query.core.dataframe.get_application_node_id")
def test_from_view_factory_error(
    get_last_application_version_mock, view, environment, api_client_obj
):
    version = "0.1.2"
    get_last_application_version_mock.return_value = version
    with responses.RequestsMock() as rsps:
        rsps.add(
            responses.GET,
            "{}/api/v1/applications/app/views?version={}".format(ORIGIN, version),
            json={"response": "ok", "data": VIEWS_DATA},
        )

        with pytest.raises(RuntimeError):
            _ = GantryDataFrame.from_view(
                api_client=api_client_obj,
                application="app",
                view=view,
                version=version,
                environment=environment,
            )



import logging
from pathlib import Path, PurePath
from typing import Any, Dict, List, Optional

from typeguard import typechecked

from gantry.api_client import APIClient
from gantry.dataset.gantry_dataset import GantryDataset
from gantry.exceptions import DatasetNotFoundError, GantryRequestException

logger = logging.getLogger(__name__)


class GantryDatasetClient:
    def __init__(
        self,
        api_client: APIClient,
        working_directory: str = str(Path().absolute()),
    ):
        self._api_client = api_client
        self._workspace = working_directory

    @property
    def working_directory(self) -> str:
        return self._workspace  # type: ignore

    @working_directory.setter
    def working_directory(self, value: str):
        if not PurePath(value).is_absolute():
            logger.warning(
                f"{value} is not an absolute path, will use {Path(value).resolve()} instead."
            )
        else:
            logger.info(f"Set dataset working directory to {value}")
        self._workspace = str(Path(value).resolve())
        Path(self._workspace).mkdir(exist_ok=True)

    def set_working_directory(self, working_dir: str) -> None:
        """
        Set the working directory for the dataset client. This is the directory where the local
        copy of the dataset will be stored.

        Example usage:

        .. code-block:: python

            import gantry.dataset as gdataset
            gdataset.set_working_directory("your working directory")

        Args:
            working_dir (str): absolute path of the directory used to store local copy of datasets.

        Returns:
            ``None``
        """
        self.working_directory = working_dir

    @typechecked
    def create_dataset(
        self, name: str, bucket_name: Optional[str] = None, app_name: Optional[str] = None
    ) -> GantryDataset:
        """
        Create a dataset with the provided name. If ``app_name`` is provided, we will use the
        application schema to set the dataset schema. ``bucket_name`` can be provided if you want
        to store data in your own bucket, please contact Gantry support if you want to use
        this feature. If not provided, we will use a gantry managed bucket.

        Example usage:

        .. code-block:: python

            import gantry.dataset as gdataset
            # Create an empty dataset with the name "dataset_name"
            gdataset.create_dataset("dataset_name")
            # Create a dataset with the name "dataset_name" and set the schema to the schema of the
            # application "app_name"
            gdataset.create_dataset("dataset_with_app", app_name="app_name")


        Args:
            name (str): dataset name
            bucket_name (str): Provide bucket name if you want to use your own bucket. If not
                provided we will use a gantry managed bucket.
            app_name (Optional[str], optional): gantry application name which will be used to set
                dataset schema if provided.

        Returns:
            :class:`gantry.dataset.GantryDataset`: an object representing the created dataset.
        """
        data = {
            "name": name,
        }

        if app_name:
            res = self._api_client.request(
                "GET", f"/api/v1/applications/{app_name}/schemas", raise_for_status=True
            )
            data["model_id"] = res["data"]["id"]

        if bucket_name:
            data.update(
                {
                    "bucket_name": bucket_name,
                }
            )

        res = self._api_client.request("POST", "/api/v1/datasets", json=data, raise_for_status=True)

        return GantryDataset(
            api_client=self._api_client,
            dataset_name=res["data"]["name"],
            dataset_id=res["data"]["id"],
            bucket_name=res["data"]["bucket_name"],
            aws_region=res["data"]["aws_region"],
            dataset_s3_prefix=f"{res['data']['organization_id']}/{res['data']['s3_prefix']}",
            workspace=self._workspace,
        )

    @typechecked
    def get_dataset(self, name: str, app_name: Optional[str] = None) -> GantryDataset:
        """
        Get a dataset object by name. If the dataset is marked as deleted, a warning will be
        logged. If the dataset does not exist, a :class:`gantry.exceptions.DatasetNotFoundError`
        will be raised.

        Example usage:

        .. code-block:: python

            import gantry.dataset as gdataset
            dataset = gdataset.get_dataset("dataset_name", "test_app")

        Args:
            name (str): the name of the dataset.
            app_name (str): the name of the application the dataset belongs to.

        Returns:
            :class:`gantry.dataset.GantryDataset`: An object representing the dataset name.
        """
        if not app_name:
            try:
                res = self._api_client.request(
                    "GET", f"/api/v1/datasets/{name}", raise_for_status=True
                )
            except GantryRequestException as e:
                if e.status_code == 404:
                    raise DatasetNotFoundError(f'Could not find dataset with name "{name}"') from e
                raise
        else:
            try:
                res = self._api_client.request(
                    "GET", f"/api/v1/applications/{app_name}/datasets/{name}", raise_for_status=True
                )
            except GantryRequestException as e:
                if e.status_code == 404:
                    raise DatasetNotFoundError(f'Could not find dataset with name "{name}"') from e
                raise

        if res["data"]["disabled"]:
            logger.warning("This dataset is marked as deleted")

        return GantryDataset(
            api_client=self._api_client,
            dataset_name=res["data"]["name"],
            dataset_id=res["data"]["id"],
            bucket_name=res["data"]["bucket_name"],
            aws_region=res["data"]["aws_region"],
            dataset_s3_prefix=f"{res['data']['organization_id']}/{res['data']['s3_prefix']}",
            workspace=self._workspace,
        )

    @typechecked
    def list_dataset_versions(self, dataset_name: str) -> List[Dict[str, Any]]:
        """
        List all versions of a dataset. Each dataset version is a snapshot of the dataset at a
        particular point in time. The result will be sorted from latest to earliest.

        Example usage:

        .. code-block:: python

            import gantry.dataset as gdataset
            gdataset.list_dataset_versions("dataset_name")

        Output example:

        .. code-block:: text

            [{'version_id': 'd6f42bd0-b7a9-4089-88fb-be7cee6d4a83',
            'dataset': 'dataset_name',
            'message': 'version info',
            'created_at': 'Wed, 08 Feb 2023 22:02:02 GMT',
            'created_by': 'db459d6d-c83b-496d-b659-e48bca971156',
            'is_latest_version': True},
            {'version_id': 'b787034a-798b-4bb3-a726-0e197ddb8aff',
            'dataset': 'dataset_name',
            'message': 'initial dataset commit',
            'created_at': 'Wed, 08 Feb 2023 22:00:26 GMT',
            'created_by': 'db459d6d-c83b-496d-b659-e48bca971156',
            'is_latest_version': False}]

        Args:
            name (str): the name of the dataset.

        Returns:
            ``List[Dict[str, Any]]``: list of dictionaries with metadata.
        """
        dataset = self.get_dataset(dataset_name)

        return dataset.list_versions()

    @typechecked
    def list_datasets(
        self, include_deleted: bool = False, model_node_id: Optional[str] = None
    ) -> List[Dict[str, Any]]:
        """
        List all datasets. If ``include_deleted`` is set to ``True``, deleted datasets will be
        included in the result list.

        Example usage:

        .. code-block:: python

            import gantry.dataset as gdataset
            gdataset.list_datasets()

        Output example:

        .. code-block:: text

            [{'name': 'dataset_0',
            'dataset_id': '44e8dff0-1e4c-4484-843c-ca3c585d405d',
            'created_at': 'Sun, 12 Feb 2023 00:08:12 GMT'},
            {'name': 'dataset_1',
            'dataset_id': '3adb66fa-9dc7-4a60-86b6-83389f567186',
            'created_at': 'Sun, 12 Feb 2023 00:05:15 GMT'},
            {'name': 'dataset_2',
            'dataset_id': '0a5b5706-2060-4601-8c5c-22900f06d54a',
            'created_at': 'Wed, 08 Feb 2023 22:00:26 GMT'},

        Args:
            include_deleted (bool): whether include deleted datasets in the result, defaults
                to ``False``.

        Returns:
            ``List[Dict[str, Any]]``: List of dictionaries, each representing one
            dataset and associated metadata.
        """
        res = self._api_client.request(
            "GET",
            "/api/v1/datasets",
            raise_for_status=True,
            params={"include_deleted": include_deleted, "model_node_id": model_node_id},
        )

        return [self._prune_dataset_info(dataset_info) for dataset_info in res["data"]]

    @typechecked
    def delete_dataset(self, name: str) -> None:
        """
        Mark the dataset as deleted. As for now this will only hide dataset from the list_datasets
        API results, it will not delete the dataset data from the server, we will release the
        hard deletion feature later.

        Example usage:

        .. code-block:: python

            import gantry.dataset as gdataset
            gdataset.delete_dataset("dataset_name")

        Args:
            name (str): the name of the dataset to be deleted.

        Returns:
            ``None``
        """
        dataset = self.get_dataset(name)

        return dataset.delete()

    def ping(self) -> bool:
        """
        Pings the API client server to check if it is alive.
        Returns ``True`` if alive, ``False`` if there is an error during ping process.

        Returns
            ``bool``: flag indicating liveness.
        """
        try:
            # Cannot use /healthz/* endpoints as those will be answered by nginx
            # need to use /.
            # See https://linear.app/gantry/issue/ENG-2978/revisit-ping-in-sdk
            self._api_client.request("GET", "/api/ping", raise_for_status=True)
            return True
        except Exception as e:
            logger.error(f"Error during ping: {e}")
            return False

    def ready(self) -> bool:
        """
        Checks whether the API is ready to receive traffic with the provided API Key.
        """
        try:
            self._api_client.request("GET", "/api/v1/auth-ping", raise_for_status=True)
            return True
        except Exception as e:
            logger.error(f"Error during api key check: {e}")
            return False

    @staticmethod
    def _prune_dataset_info(dataset_info):
        return {
            "name": dataset_info["name"],
            "dataset_id": dataset_info["id"],
            "created_at": dataset_info["created_at"],
        }



from gantry.query.core.dataframe import Filters, _FilterObj


def test_and(series_obj):
    f1 = Filters([_FilterObj({"foo": "bar"}, series_obj)])
    f2 = Filters([_FilterObj({"bar": "baz"}, series_obj)])

    f_and = f1 & f2

    assert [f.filter_ for f in f_and.filter_objs] == [{"foo": "bar"}, {"bar": "baz"}]
    assert f_and.filter_objs[0].series is series_obj
    assert f_and.filter_objs[1].series is series_obj


def test_from_single_filter(series_obj):
    filter_ = Filters._from_single_filter({"foo": "bar"}, series_obj)
    assert [f.filter_ for f in filter_.filter_objs] == [{"foo": "bar"}]
    assert filter_.filter_objs[0].series is series_obj



import mock
import pytest
import responses
from responses import matchers

from gantry import api_client
from gantry.exceptions import GantryRequestException
from gantry.version import __version__

HOST = "http://test-api"
FAKE_API_KEY = "ABCD1234"


@mock.patch("gantry.api_client.platform")
def test_request_ua(mock_platform):
    mock_platform.python_implementation.return_value = "foo"
    mock_platform.python_version.return_value = "bar"
    mock_platform.system.return_value = "baz"
    mock_platform.machine.return_value = "qux"

    api_client_obj = api_client.APIClient(HOST, FAKE_API_KEY)

    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{HOST}/health",
            headers={"Content-Type": "application/json"},
            json={
                "response": "ok",
                "data": "foobar",
            },
            match=[
                matchers.header_matcher(
                    {
                        "User-Agent": f"gantry/{__version__} foo/bar baz qux",
                        "X-Gantry-Api-Key": FAKE_API_KEY,
                        "Other": "header",
                    }
                )
            ],
        )

        assert api_client_obj.request("GET", "/health", headers={"Other": "header"}) == {
            "response": "ok",
            "data": "foobar",
        }


def test_request_rate_limit(caplog):
    api_client_obj = api_client.APIClient(HOST, FAKE_API_KEY)
    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{HOST}/health",
            headers={"Content-Type": "application/json"},
            status=429,
            json={
                "response": "ok",
                "error": "something",
            },
        )

        assert api_client_obj.request("GET", "/health") == {
            "response": "ok",
            "error": "something",
        }

        with pytest.raises(GantryRequestException) as excinfo:
            api_client_obj.request("GET", "/health", raise_for_status=True)
        assert "Logger has hit the rate limit" in str(excinfo.value)


@pytest.mark.parametrize("status_code", list(range(400, 600)))
def test_request_raise_for_status(status_code):
    api_client_obj = api_client.APIClient(HOST, FAKE_API_KEY)
    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{HOST}/some_resource",
            headers={"Content-Type": "application/json"},
            status=status_code,
            json={
                "response": "ok",
            },
        )

        with pytest.raises(GantryRequestException):
            api_client_obj.request("GET", "/some_resource", raise_for_status=True)


@pytest.mark.parametrize(
    ("status_code", "text", "expected_error"),
    [
        (404, "[error-details123]", r"A resource was not found.+\[error\-details123\]"),
        (429, "", r"Logger has hit the rate limit."),
        (401, "", r"Authentication error. Ensure that you supplied a working API key"),
        (403, "", r"Access denied. Check with your organization admin or Gantry support"),
        (409, "[error-details123]", r"Malformed data error.+\[error\-details123\]"),
        (500, "", r"Internal Gantry error. Contact Gantry support for help."),
    ],
)
def test_request_error_messages(status_code, text, expected_error):
    api_client_obj = api_client.APIClient(HOST, FAKE_API_KEY)
    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{HOST}/some_resource",
            headers={"Content-Type": "application/json"},
            status=status_code,
            json={"response": text},
        )

        with pytest.raises(GantryRequestException, match=expected_error):
            api_client_obj.request("GET", "/some_resource", raise_for_status=True)



import datetime
import re

import mock
import pytest
import responses
from responses import matchers

from gantry.api_client import APIClient
from gantry.exceptions import GantryException, GantryRequestException
from gantry.query.client import COLOR_OPTIONS_FOR_VIEW, GantryQuery
from gantry.query.core.dataframe import GantryDataFrame

from .conftest import END_TIME, ORIGIN, SOME_DURATION, START_TIME


@pytest.fixture
def gantry_query_obj():
    api_client = APIClient(origin=ORIGIN, api_key="abcd1234")
    return GantryQuery(api_client)


@pytest.mark.parametrize(
    ["data", "expected"],
    [
        ({"application_1": {}, "application_2": {}}, ["application_1", "application_2"]),
        (
            {
                "application": {},
            },
            ["application"],
        ),
        ({}, []),
    ],
)
def test_list_applications(data, expected, gantry_query_obj):
    with responses.RequestsMock() as rsps:
        rsps.add(
            responses.GET,
            "{}/api/v1/models".format(ORIGIN),
            match=[matchers.query_param_matcher({"limit": "10"})],
            json={
                "response": "ok",
                "data": data,
            },
        )

        assert gantry_query_obj.list_applications() == expected


@pytest.mark.parametrize(
    ["versions", "expected"],
    [
        ([{"internal_version": 10, "version": "baz"}], ["baz"]),
        (
            [
                {"internal_version": 10, "version": "baz"},
                {"internal_version": 20, "version": "bar"},
            ],
            ["baz", "bar"],
        ),
        ([], []),
    ],
)
def test_list_application_versions(versions, expected, gantry_query_obj):
    with responses.RequestsMock() as rsps:
        rsps.add(
            responses.GET,
            "{}/api/v1/models/foobar".format(ORIGIN),
            match=[matchers.query_param_matcher({"include_names": "True"})],
            json={"response": "ok", "data": {"versions": versions}},
        )

        assert gantry_query_obj.list_application_versions("foobar") == expected


@pytest.mark.parametrize(
    ["envs", "expected"],
    [
        (["A", "B", "C"], ["A", "B", "C"]),
        ([], []),
    ],
)
def test_list_application_environments(envs, expected, gantry_query_obj):
    with responses.RequestsMock() as rsps:
        rsps.add(
            responses.GET,
            "{}/api/v1/models/foobar".format(ORIGIN),
            json={"response": "ok", "data": {"environments": envs}},
        )

        assert gantry_query_obj.list_application_environments("foobar") == expected


@pytest.mark.parametrize("batch_id", [None, "foobar"])
@mock.patch(
    "gantry.query.client._build_batch_id_filter", return_value={"batch_id_filter": "barbaz"}
)
@mock.patch("gantry.query.client._get_application_batches", return_value=[{"id": "foobar"}])
def test_query(mock_get_application_batches, mock_build_batch_filter, batch_id, gantry_query_obj):
    application = "foobar"
    version = "1.2.3"
    environment = "env"

    with responses.RequestsMock() as rsps:
        rsps.add(
            responses.GET,
            "{}/api/v1/models/foobar/schemas".format(ORIGIN),
            json={"response": "ok", "data": {"id": "ABCD1234", "environments": []}},
            match=[
                matchers.query_param_matcher(
                    {
                        "end_time": "2008-09-10 23:58:23",
                        "start_time": "2008-09-03 20:56:35",
                        "version": "1.2.3",
                    }
                )
            ],
        )

        gantry_df = gantry_query_obj.query(
            application,
            START_TIME,
            END_TIME,
            version,
            environment,
            [{"filter": "foobar"}],
            batch_id=batch_id,
        )

        assert isinstance(gantry_df, GantryDataFrame)
        assert gantry_df.api_client == gantry_query_obj._api_client
        assert gantry_df.filters == (
            [{"filter": "foobar"}]
            if batch_id is None
            else [{"filter": "foobar"}, {"batch_id_filter": "barbaz"}]
        )

        assert gantry_df.query_info.application == application
        assert gantry_df.query_info.version == version
        assert gantry_df.query_info.environment == environment
        assert gantry_df.query_info.start_time == START_TIME
        assert gantry_df.query_info.end_time == END_TIME


@mock.patch("gantry.query.client.GantryDataFrame.from_view")
@mock.patch("gantry.query.client.GantryQuery.list_application_views")
def test_query_with_view(mock_application_views, mock_factory, gantry_query_obj):
    mock_application_views.return_value = ["barbaz"]
    mock_factory.return_value = "data"

    assert (
        gantry_query_obj.query(
            "foobar", version="1.2.3", environment="foo", view="barbaz", tags={"foo": "bar"}
        )
        == "data"
    )
    mock_factory.assert_called_once_with(
        gantry_query_obj._api_client, "foobar", "barbaz", "1.2.3", "foo", tags={"foo": "bar"}
    )


@mock.patch("gantry.query.client.GantryQuery.list_application_views")
def test_query_with_view_not_found(mock_application_views, gantry_query_obj):
    mock_application_views.return_value = ["A", "B", "C"]

    with pytest.raises(ValueError):
        _ = gantry_query_obj.query("foobar", view="not-found")


@mock.patch(
    "gantry.query.client._get_application_batches",
    return_value=[{"id": "foobar"}, {"id": "barbaz"}],
)
def test_query_with_batch_id_not_found(mock_application_batches, gantry_query_obj):
    with pytest.raises(ValueError):
        _ = gantry_query_obj.query("foobar", batch_id="not-found")


@pytest.mark.parametrize(
    ["start_time", "end_time", "filters", "batch_id"],
    [
        (START_TIME, None, None, None),
        (START_TIME, None, {}, None),
        (START_TIME, END_TIME, None, None),
        (None, END_TIME, {}, None),
        (None, END_TIME, None, None),
        (START_TIME, END_TIME, {"foo": "bar"}, None),
        (None, None, {"foo": "bar"}, None),
        (None, END_TIME, {"foo": "bar"}, None),
        (START_TIME, None, {"foo": "bar"}, None),
        (None, None, {}, "foobar"),
        (None, None, {}, None),
        (START_TIME, None, {"foo": "bar"}, "foobar"),
    ],
)
def test_query_invalid_call_with_view_and_others(
    start_time, end_time, filters, batch_id, gantry_query_obj
):
    with pytest.raises(ValueError):
        _ = gantry_query_obj.query(
            "foobar", start_time, end_time, filters=filters, batch_id=batch_id, view="foobar"
        )


def test_get_current_feedback_schema(gantry_query_obj):
    with pytest.raises(NotImplementedError):
        gantry_query_obj.get_current_metric_schema("application")


def test_update_feedback_schema(gantry_query_obj):
    with pytest.raises(NotImplementedError):
        gantry_query_obj.update_feedback_schema("application", [])


def test_add_feedback_field(gantry_query_obj):
    with pytest.raises(NotImplementedError):
        gantry_query_obj.add_feedback_field("application", {})


def test_get_current_metric_schema(gantry_query_obj):
    with pytest.raises(NotImplementedError):
        gantry_query_obj.get_current_metric_schema("application")


def test_update_metric_schema(gantry_query_obj):
    with pytest.raises(NotImplementedError):
        gantry_query_obj.update_metric_schema("application", [])


def test_add_metric(gantry_query_obj):
    with pytest.raises(NotImplementedError):
        gantry_query_obj.add_metric("application", {})


@pytest.mark.parametrize(
    ["duration", "start_time", "end_time"],
    [
        (SOME_DURATION, START_TIME, None),
        (SOME_DURATION, None, END_TIME),
        (SOME_DURATION, START_TIME, END_TIME),
        (None, None, END_TIME),
        (None, START_TIME, None),
        (None, None, None),
    ],
)
@mock.patch("gantry.query.client.get_application_node_id")
def test_create_view_invalid_time_params(
    mock_get_app_node_id, gantry_query_obj, duration, start_time, end_time
):
    mock_get_app_node_id.return_value = "foobar12345"
    with pytest.raises(ValueError):
        gantry_query_obj.create_view(
            application="foo",
            name="bar",
            version="2",
            duration=duration,
            start_time=start_time,
            end_time=end_time,
        )


@pytest.mark.parametrize(
    ["app", "version", "error_msg"],
    [
        ("foo", "2", re.escape("Application foo[2] does not exist.")),
        ("foobar", None, re.escape("Application foobar[latest] does not exist.")),
    ],
)
@mock.patch("gantry.query.client.get_application_node_id")
def test_create_view_inexistent_model(
    mock_get_app_node_id, gantry_query_obj, app, version, error_msg
):
    mock_get_app_node_id.side_effect = GantryRequestException("staging.com", 100, "error msg")
    with pytest.raises(ValueError, match=error_msg):
        gantry_query_obj.create_view(
            application=app,
            name="bar",
            version=version,
            duration=datetime.timedelta(days=1),
        )


@pytest.mark.parametrize(
    ["kwargs", "expected_data"],
    [
        (
            {
                "application": "baz",
                "name": "foo",
                "version": "1.2.3",
                "tag_filters": {"foo": "bar"},
                "data_filters": [{"some": "filter"}],
                "duration": datetime.timedelta(days=1),
            },
            {
                "name": "foo",
                "model_node_id": "foobar12345",
                "tag_filters": {"foo": "bar"},
                "data_filters": [{"some": "filter"}],
                "details": {"color": "#foobar"},
                "duration": "P1D",
            },
        ),
        (
            {
                "application": "baz",
                "name": "foo",
                "version": "1.2.3",
                "data_filters": [],
                "start_time": START_TIME,
                "end_time": END_TIME,
            },
            {
                "name": "foo",
                "model_node_id": "foobar12345",
                "tag_filters": {},
                "data_filters": [],
                "details": {"color": "#foobar"},
                "start_time": "2008-09-03T20:56:35Z",
                "end_time": "2008-09-10T23:58:23Z",
            },
        ),
    ],
)
@mock.patch("gantry.query.client.random.choice", return_value="#foobar")
@mock.patch("gantry.query.client.get_application_node_id")
def test_create_view(mock_get_app_node_id, mock_choice, kwargs, expected_data, gantry_query_obj):
    mock_get_app_node_id.return_value = "foobar12345"
    with responses.RequestsMock() as rsps:
        rsps.add(
            responses.POST,
            "{}/api/v1/views".format(ORIGIN),
            json={
                "response": "ok",
                "data": "OK",
            },
            match=[matchers.json_params_matcher(expected_data)],
        )

        gantry_query_obj.create_view(**kwargs)
    mock_get_app_node_id.assert_called_once_with(
        gantry_query_obj._api_client, kwargs["application"], version=kwargs["version"]
    )
    mock_choice.assert_called_once_with(COLOR_OPTIONS_FOR_VIEW)


@pytest.mark.parametrize(
    ["kwargs", "expected_data"],
    [
        (
            {
                "application": "baz",
                "name": "foo",
                "version": "1.2.3",
                "tag_filters": {"foo": "bar"},
                "data_filters": [{"some": "filter"}],
                "duration": datetime.timedelta(days=1),
            },
            {
                "name": "foo",
                "model_node_id": "foobar12345",
                "tag_filters": {"foo": "bar"},
                "data_filters": [{"some": "filter"}],
                "details": {"color": "#foobar"},
                "duration": "P1D",
            },
        ),
        (
            {
                "application": "baz",
                "name": "foo",
                "version": "1.2.3",
                "data_filters": [],
                "start_time": START_TIME,
                "end_time": END_TIME,
            },
            {
                "name": "foo",
                "model_node_id": "foobar12345",
                "tag_filters": {},
                "data_filters": [],
                "details": {"color": "#foobar"},
                "start_time": "2008-09-03T20:56:35Z",
                "end_time": "2008-09-10T23:58:23Z",
            },
        ),
    ],
)
@mock.patch("gantry.query.client.random.choice", return_value="#foobar")
@mock.patch("gantry.query.client.get_application_node_id")
def test_create_view_error(
    mock_get_app_node_id, mock_choice, kwargs, expected_data, gantry_query_obj
):
    mock_get_app_node_id.return_value = "foobar12345"
    with responses.RequestsMock() as rsps:
        rsps.add(
            responses.POST,
            "{}/api/v1/views".format(ORIGIN),
            json={
                "response": "not ok",
                "data": "OK",
            },
            status=400,
            match=[matchers.json_params_matcher(expected_data)],
        )

        with pytest.raises(GantryException):
            gantry_query_obj.create_view(**kwargs)

        mock_choice.assert_called_once_with(COLOR_OPTIONS_FOR_VIEW)


VIEWS_DATA = [
    {"name": "view_1", "tag_filters": {"env": "env-A"}, "other": "A"},
    {"name": "view_2", "tag_filters": {"env": "env-B"}, "other": "B"},
    {"name": "view_3", "tag_filters": {"env": "env-B"}, "other": "C"},
]


@pytest.mark.parametrize(
    ["env", "response_data", "expected", "version"],
    [("dev", [], [], "1.2.3"), ("env-A", VIEWS_DATA, ["view_1"], "1.2.3")],
)
def test_list_application_views_with_other_env(
    gantry_query_obj, env, response_data, expected, version
):
    with responses.RequestsMock() as rsps:
        rsps.add(
            responses.GET,
            "{}/api/v1/applications/barbaz/views?version={}".format(ORIGIN, version),
            json={
                "response": "ok",
                "data": response_data,
            },
        )

        assert (
            gantry_query_obj.list_application_views("barbaz", version="1.2.3", environment=env)
            == expected
        )


@pytest.mark.parametrize(
    ["response_data", "version"],
    [([], "1.2.3"), (VIEWS_DATA, "1.2.3")],
)
def test_print_application_info(response_data, version, gantry_query_obj):
    with responses.RequestsMock() as rsps:
        rsps.add(
            responses.GET,
            "{}/api/v1/models/barbaz".format(ORIGIN),
            json={
                "response": "ok",
                "data": {
                    "versions": [
                        {"internal_version": 10, "version": version},
                    ],
                    "environments": ["dev", "prod"],
                },
            },
        )
        rsps.add(
            responses.GET,
            "{}/api/v1/applications/barbaz/views".format(ORIGIN),
            json={
                "response": "ok",
                "data": response_data,
            },
            match=[matchers.query_param_matcher({"version": version})],
        )

        assert gantry_query_obj.print_application_info("barbaz") is None


@mock.patch("gantry.query.client._get_application_batches", return_value=["foo", "bar"])
def test_list_application_batches(mock_get_application_batches, gantry_query_obj):
    assert gantry_query_obj.list_application_batches("app", "2") == ["foo", "bar"]
    mock_get_application_batches.assert_called_once_with(gantry_query_obj._api_client, "app", "2")



from datetime import datetime

import mock
import numpy as np
import pandas as pd
import pytest
import responses
from freezegun import freeze_time
from responses import matchers

from gantry.query.core.utils import (
    _build_batch_id_filter,
    _build_empty_df_with_schema,
    _get_application_batches,
    get_application_node_id,
    get_application_views,
    get_last_application_version,
    get_start_end_time_from_view,
    runs_on,
    same_size,
)

from ..conftest import DATA, END_TIME, END_TIME_STR, ORIGIN, START_TIME, START_TIME_STR
from tests.query.test_client import VIEWS_DATA


@pytest.mark.parametrize(
    ["supported", "invalid"],
    [
        (["bool"], "int"),
        (["bool"], "object"),
        (["bool"], "float"),
        (["bool", "float"], "int"),
        (["int", "str"], "bool"),
        (["int", "bool"], "str"),
    ],
)
def test_runs_on_invalid(supported, invalid):
    dec = runs_on(*supported)
    f = dec(lambda m: True)
    with pytest.raises(ValueError):
        f(mock.Mock(dtype=invalid))


@pytest.mark.parametrize(
    ["supported", "valid"],
    [
        (["bool"], "bool"),
        (["bool", "int"], "int"),
        (["str", "float"], "float"),
        (["str", "float"], "str"),
    ],
)
def test_runs_on_valid(supported, valid):
    dec = runs_on(*supported)
    f = dec(lambda m: True)
    assert f(mock.Mock(dtype=valid))


l_0 = []
l_1 = list(range(1))
l_10 = list(range(10))
l_100 = list(range(100))


@pytest.mark.parametrize("allow_empty", [False, True])
@pytest.mark.parametrize(
    ["args", "kwargs"],
    [
        ([l_10, l_10], {"s3": l_10}),
        ([l_10], {"s3": l_10, "s2": l_10}),
        ([], {"s1": l_10, "s3": l_10, "s2": l_10}),
    ],
)
def test_same_size_valid(args, kwargs, allow_empty):
    dec = same_size("s1", "s2", "s3", allow_empty=allow_empty)
    f = dec(lambda s1, s2, s3: True)
    assert f(*args, **kwargs)


@pytest.mark.parametrize(
    ["args", "kwargs"],
    [
        ([l_0, l_0], {"s3": l_0}),
        ([l_0], {"s3": l_0, "s2": l_0}),
        ([], {"s1": l_0, "s3": l_0, "s2": l_0}),
    ],
)
def test_same_size_valid_allow_empty(args, kwargs):
    dec = same_size("s1", "s2", "s3", allow_empty=True)
    f = dec(lambda s1, s2, s3: True)
    assert f(*args, **kwargs)


@pytest.mark.parametrize("allow_empty_kwarg", [{}, {"allow_empty": False}])  # Test default value
@pytest.mark.parametrize(
    ["args", "kwargs"],
    [
        ([l_0, l_0], {"s3": l_0}),
        ([l_0], {"s3": l_0, "s2": l_0}),
        ([], {"s1": l_0, "s3": l_0, "s2": l_0}),
    ],
)
def test_same_size_valid_not_allow_empty(args, kwargs, allow_empty_kwarg):
    dec = same_size("s1", "s2", "s3", **allow_empty_kwarg)
    f = dec(lambda s1, s2, s3: True)
    with pytest.raises(ValueError):
        f(*args, **kwargs)


@pytest.mark.parametrize("allow_empty", [False, True])
@pytest.mark.parametrize(
    ["args", "kwargs"],
    [
        ([l_10, l_100], {"s3": l_10}),
        ([l_10], {"s3": l_100, "s2": l_100}),
        ([l_0], {"s3": l_100, "s2": l_100}),
        ([], {"s1": l_10, "s3": l_100, "s2": l_10}),
        ([], {"s1": l_10, "s3": l_100, "s2": l_1}),
        ([], {"s1": l_0, "s3": l_100, "s2": l_10}),
        ([], {"s1": l_0, "s3": l_100, "s2": l_0}),
    ],
)
def test_same_size_not_same_size_error(args, kwargs, allow_empty):
    dec = same_size("s1", "s2", "s3", allow_empty=allow_empty)
    f = dec(lambda s1, s2, s3: True)
    with pytest.raises(ValueError):
        f(*args, **kwargs)


@pytest.mark.parametrize("bad_params", [["other"], ["s1", "s2", "s3"]])
def test_same_size_bad_params(bad_params):
    dec = same_size(*bad_params)
    with pytest.raises(ValueError):
        _ = dec(lambda s1, s2: True)


@pytest.mark.parametrize(
    ["start_time", "end_time", "version", "content"],
    [
        (
            START_TIME,
            END_TIME,
            "1.2.3",
            {
                "end_time": "2008-09-10 23:58:23",
                "start_time": "2008-09-03 20:56:35",
                "version": "1.2.3",
            },
        ),
        (None, None, None, {}),
        (None, None, "2", {"version": "2"}),
        (
            START_TIME,
            END_TIME,
            None,
            {
                "end_time": "2008-09-10 23:58:23",
                "start_time": "2008-09-03 20:56:35",
            },
        ),
    ],
)
def test_get_application_node_id(start_time, end_time, version, content, api_client_obj):
    with responses.RequestsMock() as rsps:
        rsps.add(
            responses.GET,
            "{}/api/v1/models/foobar/schemas".format(ORIGIN),
            json={"response": "ok", "data": DATA},
            match=[matchers.query_param_matcher(content)],
        )

        assert (
            get_application_node_id(api_client_obj, "foobar", start_time, end_time, version)
            == "ABCD1234"
        )


@pytest.mark.parametrize(
    ["start_time", "end_time", "version", "content"],
    [
        (
            START_TIME,
            END_TIME,
            "1.2.3",
            {
                "end_time": "2008-09-10 23:58:23",
                "start_time": "2008-09-03 20:56:35",
                "version": "1.2.3",
            },
        ),
        (None, None, None, {}),
        (None, None, "2", {"version": "2"}),
        (
            START_TIME,
            END_TIME,
            None,
            {
                "end_time": "2008-09-10 23:58:23",
                "start_time": "2008-09-03 20:56:35",
            },
        ),
    ],
)
@pytest.mark.parametrize("json_data", [{"response": "ok"}, {"response": "ok", "data": {}}])
def test_get_application_node_id_error(
    json_data, start_time, end_time, version, content, api_client_obj
):
    with responses.RequestsMock() as rsps:
        rsps.add(
            responses.GET,
            "{}/api/v1/models/foobar/schemas".format(ORIGIN),
            json=json_data,
            match=[matchers.query_param_matcher(content)],
        )

        with pytest.raises(RuntimeError):
            _ = get_application_node_id(api_client_obj, "foobar", start_time, end_time, version)


@pytest.mark.parametrize("version", ["2", None])
@pytest.mark.parametrize(
    ["env", "expected"],
    [
        (
            "env-A",
            [{"name": "view_1", "tag_filters": {"env": "env-A"}, "other": "A"}],
        ),
        (
            "env-B",
            [
                {"name": "view_2", "tag_filters": {"env": "env-B"}, "other": "B"},
                {"name": "view_3", "tag_filters": {"env": "env-B"}, "other": "C"},
            ],
        ),
        (None, VIEWS_DATA),
    ],
)
def test_get_application_views(env, expected, version, api_client_obj):
    with responses.RequestsMock() as rsps:
        latest_version = "2"
        rsps.add(
            responses.GET,
            "{}/api/v1/applications/barbaz/views".format(ORIGIN),
            json={"response": "ok", "data": VIEWS_DATA},
            match=[matchers.query_param_matcher({"version": latest_version})],
        )

        if not version:
            rsps.add(
                responses.GET,
                "{}/api/v1/models/barbaz/schemas".format(ORIGIN),
                json={"response": "ok", "data": {"version": latest_version}},
            )

        assert (
            get_application_views(api_client_obj, "barbaz", version=version, environment=env)
            == expected
        )


@pytest.mark.parametrize("version", [None])
def test_get_last_application_version_with_error(version, api_client_obj):
    with pytest.raises(RuntimeError):
        with responses.RequestsMock() as rsps:
            rsps.add(
                responses.GET,
                "{}/api/v1/models/barbaz/schemas".format(ORIGIN),
                json={"response": "ok", "data": {"version": version}},
            )

            get_last_application_version(api_client_obj, "barbaz")


DATE_PARAM_FMT = "%Y-%m-%d+%H%%3A%M%%3A%S"


@pytest.mark.parametrize(["version"], ["1"])
@pytest.mark.parametrize(
    ["params"],
    [
        [{"start_time": None, "end_time": None}],
        [{"start_time": START_TIME.strftime(DATE_PARAM_FMT), "end_time": None}],
        [{"start_time": None, "end_time": END_TIME.strftime(DATE_PARAM_FMT)}],
        [
            {
                "start_time": START_TIME.strftime(DATE_PARAM_FMT),
                "end_time": END_TIME.strftime(DATE_PARAM_FMT),
            }
        ],
    ],
)
def test_get_last_application_version(version, params, api_client_obj):
    with responses.RequestsMock() as rsps:
        rsps.add(
            responses.GET,
            "{}/api/v1/models/barbaz/schemas".format(ORIGIN),
            json={"response": "ok", "data": {"version": version}},
            match=[matchers.query_param_matcher({k: v for k, v in params.items() if v})],
        )

        assert get_last_application_version(api_client_obj, "barbaz", **params) == version


@freeze_time("2012-01-14")
@pytest.mark.parametrize(
    ["view", "expected"],
    [
        (
            {
                "start_time": START_TIME_STR,
                "end_time": END_TIME_STR,
                "duration": "P1D",
            },
            (START_TIME, END_TIME),
        ),
        (
            {
                "start_time": None,
                "end_time": END_TIME_STR,
                "duration": "P1D",
            },
            (datetime(2012, 1, 14, 0, 0), datetime(2012, 1, 13, 0, 0)),
        ),
        (
            {
                "start_time": None,
                "end_time": END_TIME_STR,
                "duration": "P8DT10H24M32S",
            },
            (datetime(2012, 1, 14, 0, 0), datetime(2012, 1, 5, 13, 35, 28)),
        ),
    ],
)
def test_get_start_end_time_from_view(view, expected):
    assert get_start_end_time_from_view(view) == expected


@pytest.mark.parametrize(
    "invalid_view",
    [
        {},
        {
            "start_date": START_TIME_STR,
            "end_date": END_TIME_STR,
        },
    ],
)
def test_get_start_end_time_from_view_error(invalid_view):
    with pytest.raises(ValueError):
        get_start_end_time_from_view(invalid_view)


def test_build_empty_df_with_schema():
    schema = pd.Series(
        ["int", "str", "float", "category", "bool", "unk"], index=["A", "B", "C", "D", "E", "F"]
    )
    empty_df = _build_empty_df_with_schema(schema)

    assert len(empty_df) == 0
    assert len(empty_df.columns) == 6

    assert empty_df.columns[0] == "A"
    assert empty_df.columns[1] == "B"
    assert empty_df.columns[2] == "C"
    assert empty_df.columns[3] == "D"
    assert empty_df.columns[4] == "E"
    assert empty_df.columns[5] == "F"

    assert list(empty_df.dtypes.items()) == [
        ("A", np.dtype("int64")),
        ("B", np.dtype("O")),
        ("C", np.dtype("float")),
        ("D", pd.CategoricalDtype(categories=[])),
        ("E", np.dtype("bool")),
        ("F", np.dtype("O")),
    ]


def test_build_batch_id_filter():
    assert _build_batch_id_filter("foobar") == {
        "feature_name": "batch_id",
        "string_query": "foobar",
        "dtype": "reserved_field",
    }


@pytest.mark.parametrize("version", ["2", None])
@mock.patch("gantry.query.core.utils.get_application_node_id", return_value="ABCD1234")
@mock.patch("gantry.query.core.utils.get_last_application_version", return_value="3")
def test_get_application_batches(
    mock_get_last_application_version, mock_get_application_node_id, version, api_client_obj
):
    with responses.RequestsMock() as rsps:
        rsps.add(
            responses.GET,
            "{}/api/v1/applications/ABCD1234/batches".format(ORIGIN),
            json={"response": "ok", "data": {"batches": [1, 2, 3, 4, 5]}},
        )

        assert _get_application_batches(api_client_obj, "barbaz", version=version) == [
            1,
            2,
            3,
            4,
            5,
        ]

        mock_get_application_node_id.assert_called_once_with(
            api_client_obj, "barbaz", version="2" if version else "3"
        )



TIMEZONE = {"TO_TIMEZONE": "UTC"}

SCHEMA_TYPES = [
    "prediction_datanodes",
    "feedback_datanodes",
    "projection_datanodes",
]

MAX_SIZE_CACHE_B = 10 * (2**20)  # 10 MB
MAX_TTL_CACHE_SECS = 600  # 10 minutes



import re
import uuid
from typing import Optional, Tuple
from urllib.parse import urlparse

import backoff
import boto3
import botocore
import click
import requests
from boto3 import Session
from click_aliases import ClickAliasedGroup
from halo import Halo
from label_studio_sdk import Client as LSClient  # type: ignore
from label_studio_sdk import Project  # type: ignore
from requests.exceptions import HTTPError
from tabulate import tabulate

import gantry
from gantry import dataset as gdataset
from gantry import query as gquery
from gantry.api_client import APIClient
from gantry.exceptions import DatasetNotFoundError, GantryRequestException


class LabelStudioManager:
    def __init__(self, ls_cli: LSClient) -> None:
        self.ls_cli = ls_cli

    def check_status(self):
        status = self.ls_cli.check_connection().get("status")
        if status != "UP":
            raise RuntimeError(f"LS instance not accessible: {status}")

        try:
            self.ls_cli.get_projects()
        except HTTPError as err:
            if err.response.status_code == 401:
                raise RuntimeError("Invalid Label Studio user token")
            else:
                raise RuntimeError(f"Cannot access LS: {err.response.text}")

    def create_project(self, name: str, project_interface: str):
        with open(project_interface, "r") as f:
            return self.ls_cli.start_project(
                title=name, description="Gantry generated project", label_config=f.read()
            )


def _register_import_storage(
    project: Project,
    bucket: str,
    source_data_folder: str,
    s3_access_key: str,
    s3_secret_key: str,
    s3_token: str,
) -> str:
    # https://labelstud.io/sdk/project.html#label_studio_sdk.project.Project.connect_s3_import_storage
    s3_storage = project.connect_s3_import_storage(
        bucket=bucket,
        prefix=source_data_folder,
        use_blob_urls=False,
        title="Gantry S3 connection",
        description="Integration made by Gantry",
        aws_access_key_id=s3_access_key,
        aws_secret_access_key=s3_secret_key,
        aws_session_token=s3_token,
    )
    return s3_storage["id"]


def _register_export_storage(
    project: Project,
    bucket: str,
    source_data_folder: str,
    s3_access_key: str,
    s3_secret_key: str,
    s3_token,
) -> str:
    # https://labelstud.io/sdk/project.html#label_studio_sdk.project.Project.connect_s3_export_storage
    s3_storage = project.connect_s3_export_storage(
        bucket=bucket,
        prefix=source_data_folder,
        title="Gantry S3 connection",
        description="Integration made by Gantry",
        aws_access_key_id=s3_access_key,
        aws_secret_access_key=s3_secret_key,
        aws_session_token=s3_token,
    )
    return s3_storage["id"]


def _sync_import_storage(storage_id: str, ls_host: str, ls_token: str) -> None:
    resp = requests.post(
        f"{ls_host}/api/storages/s3/{storage_id}/sync",
        headers={"Authorization": f"Token {ls_token}"},
    )
    try:
        resp.raise_for_status()
    except requests.exceptions.HTTPError as exc:
        raise RuntimeError(f"Cannot sync storage {storage_id}: {exc}. Are AWS creds correct?")


def _sync_export_storage(storage_id: str, ls_host: str, ls_token: str) -> None:
    resp = requests.post(
        f"{ls_host}/api/storages/export/s3/{storage_id}/sync",
        headers={"Authorization": f"Token {ls_token}"},
    )
    try:
        resp.raise_for_status()
    except requests.exceptions.HTTPError as e:
        raise RuntimeError(f"Cannot sync storage {storage_id}: {e}")


def _generate_s3_prefix(name) -> str:
    id_ = uuid.uuid4()
    return f"{name}__gantry_{str(id_)[:5]}"


def _create_folder_structure(
    s3_cli: "botocore.client.S3", bucket: str, s3_prefix: str
) -> Tuple[str, str]:
    # LS expects folder names to be without trailing slash
    folders = (f"{s3_prefix}/source_data", f"{s3_prefix}/labeled_data")

    for folder in folders:
        s3_cli.put_object(Bucket=bucket, Key=f"{folder}/")  # s3 does need the trailing slash

    return folders


@click.group(cls=ClickAliasedGroup)
def labeling():
    """
    Labeling cli
    """


CREATE_SOURCE_HELP = """
Source needs to be a URN. Currently only datasets are valid URNs:
'gantry://datasets/<dataset_name>/versions/<version_id>'.
"""


def _get_aws_creds(api_client: APIClient, bucket_name: str) -> Tuple[str, str, str, str]:
    session = Session()
    credentials = session.get_credentials()
    # Credentials are refreshable, so accessing your access key / secret key
    # separately can lead to a race condition. Use this to get an actual matched
    # set.
    current_credentials = credentials.get_frozen_credentials()

    return (
        current_credentials.access_key,
        current_credentials.secret_key,
        current_credentials.token,
        session.region_name,
    )


DATASET_URI_REGEX = re.compile(
    r"gantry://datasets/(?P<dataset_name>[a-zA-Z0-9_-]+)/"
    r"versions/(?P<dataset_version>[a-zA-Z0-9_-]+)"
)
BATCH_ID_URI_REGEX = re.compile(
    r"gantry://apps/(?P<app_name>[a-zA-Z0-9._-]+)/batches/(?P<batch_id>[a-zA-Z0-9_-]+)"
)


def _validate_dataset(name: str, version: str) -> None:
    try:
        versions = gdataset.list_dataset_versions(name)
        if version not in {version["version_id"] for version in versions}:
            raise click.BadParameter(f"Version '{version}' doesn't exist for dataset '{name}'")
    except DatasetNotFoundError:
        raise click.BadParameter(f"Dataset '{name}' doesn't exist")


def _validate_batch_id(app: str, batch_id: str) -> None:
    try:
        batches = gquery.list_application_batches(application=app)
        if batch_id not in (b["id"] for b in batches):
            raise click.BadParameter(f"Batch '{batch_id}' doesn't exist for app '{app}'")
    except GantryRequestException:
        raise click.BadParameter(f"App '{app}' doesn't exist")


def _validate_source(ctx, param, value):
    gantry.init(api_key=ctx.params["api_key"])

    dataset_regex_match = DATASET_URI_REGEX.match(value)
    batch_id_regex_match = BATCH_ID_URI_REGEX.match(value)
    if dataset_regex_match:
        dataset_name, version = (
            dataset_regex_match.groupdict()["dataset_name"],
            dataset_regex_match.groupdict()["dataset_version"],
        )
        _validate_dataset(dataset_name, version)
    elif batch_id_regex_match:
        app, batch_id = (
            batch_id_regex_match.groupdict()["app_name"],
            batch_id_regex_match.groupdict()["batch_id"],
        )
        _validate_batch_id(app, batch_id)
    else:
        raise click.BadParameter(
            f"'{value}' is not a valid source. Only "
            "'gantry://datasets/<dataset_name>/versions/<dataset_version_id>'"
            " and 'gantry://apps/<app_name>/batches/<batch_id>' URIs supported currently."
        )

    return value


def _validate_protocol(ctx, param, value):
    parsed_url = urlparse(value)
    if parsed_url.scheme == "http":
        click.secho(
            "--> WARNING: Label Studio instance over plain HTTP. Connection is not safe.",
            fg="yellow",
        )
    return value


@labeling.command(
    aliases=["create"],
    help="Create a labeling job.",
)
@click.option(
    "-s",
    "--source",
    type=click.STRING,
    required=True,
    help=CREATE_SOURCE_HELP,
    callback=_validate_source,
)
@click.option(
    "-n", "--name", type=click.STRING, required=True, help="Name of the Label studio project"
)
@click.option(
    "--bucket",
    type=click.STRING,
    required=True,
    help="S3 bucket name Gantry will use as LS backend storage",
)
@click.option("--ls-token", type=click.STRING, required=True, help="Label studio user token")
@click.option(
    "--ls-host",
    type=click.STRING,
    required=True,
    help="Label studio host",
    callback=_validate_protocol,
)
@click.option(
    "--project-interface",
    type=click.STRING,
    required=True,
    help="Path to Label studio labeling spec",
)
@click.option(
    "--api-key",
    type=click.STRING,
    required=False,
    help="Gantry api key. If not provided, GANTRY_API_KEY env var will be used",
    is_eager=True,  # need this to be eager to be accessible in source validation.
)
def create(
    source: str,
    name: str,
    bucket: str,
    ls_token: str,
    ls_host: str,
    project_interface: str,
    api_key: Optional[str],
):
    cli = LSClient(url=ls_host, api_key=ls_token)
    cli = LabelStudioManager(cli)

    api_client: APIClient = gantry.get_client().log_store._api_client
    try:
        _ = api_client.request(
            "GET",
            f"/api/v1/customer-resources/buckets/{bucket}?storage_type=S3",
            raise_for_status=True,
        )
    except GantryRequestException:
        raise RuntimeError(
            # TODO -> insert docs
            f"Bucket '{bucket}' was not registered in Gantry. Labeling "
            "requires customers to register bucket with their secrets prior "
            "to creating a labeling project. Please refer to docs on how "
            "to register a bucket."
        )

    try:
        _ = api_client.request("GET", f"/api/v1/labeling/projects/{name}", raise_for_status=True)
        raise ValueError(
            f"Name must be unique. A labeling project with name '{name}' already exists"
        )
    except GantryRequestException as exc:
        if exc.status_code != 404:
            raise

    cli.check_status()
    click.secho("--> Label Studio instance accessible.", fg="green")

    project = cli.create_project(name, project_interface)
    click.secho(f"--> New Label Studio project created [id={project.id}].", fg="green")

    aws_access_key_id, aws_secret_access_key, aws_session_token, region = _get_aws_creds(
        api_client, bucket
    )
    s3_cli = boto3.client(
        "s3",
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        aws_session_token=aws_session_token,
        region_name=region,
    )

    s3_prefix = _generate_s3_prefix(name)
    source_data_folder, labeled_data_folder = _create_folder_structure(s3_cli, bucket, s3_prefix)
    click.secho("--> S3 bucket folder creation done.", fg="green")

    import_storage_id = _register_import_storage(
        project,
        bucket,
        source_data_folder,
        aws_access_key_id,
        aws_secret_access_key,
        aws_session_token,
    )
    export_storage_id = _register_export_storage(
        project,
        bucket,
        labeled_data_folder,
        aws_access_key_id,
        aws_secret_access_key,
        aws_session_token,
    )
    click.secho(
        "--> S3 source integrated with Label Studio project.",
        fg="green",
    )

    resp = api_client.request(
        "POST",
        "/api/v1/labeling/projects",
        raise_for_status=True,
        json={
            "name": name,
            "bucket": bucket,
            "s3_prefix": s3_prefix,
            "source": source,
            "ls_host": ls_host,
            "import_storage_id": import_storage_id,
            "export_storage_id": export_storage_id,
        },
    )
    data = resp["data"]
    click.secho(
        f"--> Registered labeling project in Gantry [id={data['id']}]",
        fg="green",
    )

    with Halo(text="Pushing data into LabelStudio (this could take a while)...", spinner="dots"):
        # https://farazdagi.com/posts/2014-10-16-rest-long-running-jobs/
        resp = api_client.request(
            "PATCH",
            f"/api/v1/labeling/projects/{name}",
            raise_for_status=True,
            json={"new-status": "imported"},
            raw_response=True,
        )
        status_url = resp.headers["Location"]
        status = wait_until_success(api_client, status_url)
        if status == "FAILURE":
            click.secho(
                "--> Pushing data failed. Contact Gantry for further instruction "
                f"[project={name}]",
                fg="red",
            )
            return

        _sync_import_storage(import_storage_id, ls_host, ls_token)

    click.secho(
        (
            "--> Synced source data. Ready to start labeling: "
            f"{ls_host.rstrip('/')}/projects/{project.id}"
        ),
        fg="green",
    )


@backoff.on_predicate(backoff.constant, lambda x: x not in ["SUCCESS", "FAILURE"], interval=1)
def wait_until_success(api_client, status_url):
    resp = api_client.request("GET", status_url)
    return resp["status"]


@labeling.command(
    aliases=["list"],
    help="List all labeling job.",
)
@click.option(
    "--api-key",
    type=click.STRING,
    required=False,
    help="Gantry api key. If not provided, GANTRY_API_KEY env var will be used",
)
def list(api_key):
    gantry.init(api_key=api_key)
    api_client: APIClient = gantry.get_client().log_store._api_client
    resp = api_client.request("GET", "/api/v1/labeling/projects", raise_for_status=True)
    data = resp["data"]
    filtered_data = [
        {
            k: v
            for k, v in item.items()
            if k not in ["org_id", "id", "ls_host", "import_storage_id", "export_storage_id"]
        }
        for item in data
    ]
    if filtered_data:
        print(tabulate(filtered_data, headers="keys"))
    else:
        click.secho("--> No labeling projects found.", fg="green")


@labeling.command(
    aliases=["export"],
    help="Export labeled data from Label Studio into a Gantry dataset",
)
@click.option(
    "-n", "--name", type=click.STRING, required=True, help="Name of the Label studio project"
)
@click.option(
    "-d",
    "--destination",
    type=click.STRING,
    required=True,
    help="Name of the destination dataset. If dataset does not exist, Gantry will create it.",
)
@click.option("--ls-token", type=click.STRING, required=True, help="Label studio user token")
@click.option(
    "--api-key",
    type=click.STRING,
    required=False,
    help="Gantry api key. If not provided, GANTRY_API_KEY env var will be used",
)
def dataset_export(api_key: str, destination: str, ls_token: str, name: str) -> None:
    gantry.init(api_key=api_key)
    api_client: APIClient = gantry.get_client().log_store._api_client
    resp = api_client.request("GET", f"/api/v1/labeling/projects/{name}", raise_for_status=True)
    data = resp["data"]
    _sync_export_storage(data["export_storage_id"], data["ls_host"], ls_token)
    click.secho("--> Synced labeled data to S3 bucket", fg="green")

    new_dataset = False

    with Halo(text="Exporting labeled data...", spinner="dots"):
        try:
            dataset = gdataset.get_dataset(destination)
        except DatasetNotFoundError:
            dataset = gdataset.create_dataset(destination)
            new_dataset = True

        # https://farazdagi.com/posts/2014-10-16-rest-long-running-jobs/
        resp = api_client.request(
            "POST",
            f"/api/v1/datasets/{dataset._dataset_id}/labeled-data",
            json={
                "labeling-project-name": name,
            },
            raw_response=True,
            raise_for_status=True,
        )
        status_url = resp.headers["Location"]
        status = wait_until_success(api_client, status_url)
        if status == "FAILURE":
            click.secho(
                "--> Exporting data failed. Contact Gantry for further instruction "
                f"[project={name}]",
                fg="red",
            )
            return

        resp = api_client.request("GET", status_url)
        dataset_uri = f"gantry://datasets/{destination}/versions/{resp['id']}"

    if new_dataset:
        click.secho(f"--> Created new dataset '{destination}'", fg="yellow")
    click.secho(
        f"--> Labeled data has been successfully exported: {dataset_uri}",
        fg="green",
    )


@labeling.command(
    aliases=["export"],
    help="Export labeled data from Label Studio into Gantry dashboard.",
)
@click.option(
    "-n", "--name", type=click.STRING, required=True, help="Name of the Label studio project"
)
@click.option(
    "-d",
    "--destination",
    type=click.STRING,
    required=True,
    help="Name of the gantry application. Application needs to exist and the source data "
    "needs to have been extracted from this same application",
)
@click.option("--ls-token", type=click.STRING, required=True, help="Label studio user token")
@click.option(
    "--api-key",
    type=click.STRING,
    required=False,
    help="Gantry api key. If not provided, GANTRY_API_KEY environment variable will be used",
)
def dashboard_export(api_key: str, destination: str, ls_token: str, name: str) -> None:
    gantry.init(api_key=api_key)
    api_client: APIClient = gantry.get_client().log_store._api_client
    resp = api_client.request("GET", f"/api/v1/labeling/projects/{name}", raise_for_status=True)
    data = resp["data"]
    try:
        _ = gquery.list_application_versions(destination)
    except GantryRequestException:
        raise ValueError(
            f"Application '{destination}' was not found. Exporting to dashboard only "
            "works if application exists and source data from the labeling project "
            "was extracted from this application."
        )

    _sync_export_storage(data["export_storage_id"], data["ls_host"], ls_token)
    click.secho("--> Synced labeled data to S3 bucket", fg="green")

    with Halo(text="Exporting labeled data...", spinner="dots"):
        # https://farazdagi.com/posts/2014-10-16-rest-long-running-jobs/
        resp = api_client.request(
            "POST",
            "/api/v1/ingest/labeling",
            json={
                "application": destination,
                "labeling_project": name,
            },
            raw_response=True,
            raise_for_status=True,
        )
        status_url = resp.headers["Location"]
        status = wait_until_success(api_client, status_url)
        if status == "FAILURE":
            click.secho(
                "--> Exporting data failed. Contact Gantry for further instruction "
                f"[project={name}]",
                fg="red",
            )
            return
        resp = api_client.request("GET", status_url)
        dashboard_uri = f"https://app.gantry.io/applications/{destination}/data"

    click.secho(
        f"--> Labeled data has been successfully exported: {dashboard_uri}",
        fg="green",
    )



from gantry.query.metric.main import (  # # noqa: F401
    accuracy_score,
    confusion_matrix,
    f1_score,
    mean_squared_error,
    percent_false,
    percent_false_not_null,
    percent_null,
    percent_true,
    percent_true_not_null,
    precision_score,
    r2_score,
    recall_score,
    roc_auc_score,
)

__all__ = [
    "accuracy_score",
    "mean_squared_error",
    "confusion_matrix",
    "f1_score",
    "r2_score",
    "precision_score",
    "recall_score",
    "roc_auc_score",
    "percent_null",
    "percent_false",
    "percent_true",
    "percent_true_not_null",
    "percent_false_not_null",
]



import datetime
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Union

from gantry.exceptions import GantryBatchCreationException
from gantry.logger.constants import (
    BatchType,
    ScheduleFrequency,
    ScheduleType,
    UploadFileType,
)


@dataclass
class DataLinkElement:
    ref: Optional[int] = None
    val: Optional[Any] = None  # Exact value of the item.

    def __post_init__(self):
        if self.ref is None and self.val is None:
            raise GantryBatchCreationException("ref or val must be populated.")
        return self


@dataclass
class DataLink:
    file_type: UploadFileType
    batch_type: BatchType
    num_events: int
    application: str
    version: Optional[str]
    batch_id: Optional[str] = None
    log_timestamp: Optional[str] = None
    run_name: Optional[str] = None
    run_tags: Optional[Dict] = None
    timestamp: Dict[str, DataLinkElement] = field(default_factory=dict)
    inputs: Dict[str, DataLinkElement] = field(default_factory=dict)
    outputs: Dict[str, DataLinkElement] = field(default_factory=dict)
    feedback: Dict[str, DataLinkElement] = field(default_factory=dict)
    tags: Dict[str, DataLinkElement] = field(default_factory=dict)
    feedback_id: Dict[str, DataLinkElement] = field(default_factory=dict)
    feedback_keys: List[str] = field(default_factory=list)


@dataclass
class ScheduleOptions:
    """
    Dictionary of key and value pairs to specify the additional specifications
    for logger schedule

    delay_time: Number of seconds to delay the scheduled ingestion. Default is 0s.
    watermark_key: Column name to use as the watermark timestamp for INCREMENTAL APPEND
    """

    delay_time: Optional[int] = 0
    watermark_key: Optional[str] = None


@dataclass
class Schedule:
    """
    start_on: (Union[datetime, str]): ISO 8601 formatted string or
        datetime object representing the start time of the scheduled ingestion.
    frequency: (ScheduleFrequency): 1 hour | 2 hours | 4 hours | 8 hours |
        12 hours | 1 day
    type: (ScheduleType): INCREMENTAL_APPEND | FULL_REFRESH_APPEND
    options: (ScheduleOptions): Dictionary of key and value pairs to specify the
        additional specifications for the schedule.

        For instance, schedule_type of INCREMENTAL_APPEND requires a "watermark_key",
        which is the timestamp column name to use as the watermark to divide the
        tumbling window for each incremental append.
    """

    frequency: ScheduleFrequency
    start_on: Optional[Union[datetime.datetime, str]] = None
    type: Optional[ScheduleType] = None
    options: Optional[ScheduleOptions] = None


@dataclass
class IngestFromDataConnectorRequest:
    """
    Request body for creating a new ingestion pipeline from a data connector.
    """

    application: str
    source_data_connector_name: str
    timestamp: Optional[str]
    inputs: Optional[List[str]]
    outputs: Optional[List[str]]
    feedbacks: Optional[List[str]]
    join_key: Optional[str]
    row_tags: Optional[List[str]]
    global_tags: Optional[Dict[str, str]]
    schedule: Optional[Schedule]
    pipeline_name: Optional[str]


@dataclass
class PipelineSourceOptions:
    """
    Pipeline source options for a data connector ingestion pipeline.
    """

    inputs: Optional[List[str]]
    outputs: Optional[List[str]]
    feedbacks: Optional[List[str]]
    join_key: Optional[str]
    timestamp: Optional[str]
    global_tags: Optional[Dict[str, str]]
    row_level_tags: Optional[List[str]]


@dataclass
class IngestFromDataConnectorResponse:
    """
    Response received after creating a new ingestion pipeline from a data connector.
    """

    id: str
    name: str
    source_options: PipelineSourceOptions
    updated_at: str
    created_at: str



from gantry.automations.curators.curators import Curator  # noqa: F401
from gantry.automations.curators.main import (
    get_all_curators,
    get_curator,
    list_curators,
)
from gantry.automations.curators.stock_curators import (  # noqa: F401
    AscendedSortCurator,
    BalancedStratificationCurator,
    BoundedRangeCurator,
    DescendedSortCurator,
    EqualityCurator,
    NewestCurator,
    OldestCurator,
    ProportionalStratificationCurator,
    StrictStratificationCurator,
    UniformCurator,
)

__all__ = [
    "get_curator",
    "get_all_curators",
    "list_curators",
    "Curator",
    "AscendedSortCurator",
    "BalancedStratificationCurator",
    "BoundedRangeCurator",
    "DescendedSortCurator",
    "NewestCurator",
    "OldestCurator",
    "ProportionalStratificationCurator",
    "StrictStratificationCurator",
    "UniformCurator",
]
